#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.geometry.materials import (
    read_voxel_materials,
)
from opengate.tests import utility
from scipy.spatial.transform import Rotation

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test009_voxels", "test009")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.output_dir = paths.output / __file__.rstrip(".py")
    sim.g4_verbose = True
    sim.visu = True
    sim.visu_type = "qt"
    sim.number_of_threads = 1
    sim.output_dir = paths.output
    sim.store_json_archive = True
    sim.json_archive_filename = "simulation_test009_voxels.json"

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]
    # fake.rotation = Rotation.from_euler("x", 20, degrees=True).as_matrix()

    # image
    patient = sim.add_volume("Image", "patient")
    patient.image = paths.data / "patient-4mm.mhd"
    patient.mother = "fake"
    patient.material = "G4_AIR"  # material used by default
    patient.color = [1, 0, 1, 1]
    patient.voxel_materials = [
        [-2000, -900, "G4_AIR"],
        [-900, -100, "Lung"],
        [-100, 0, "G4_ADIPOSE_TISSUE_ICRP"],
        [0, 300, "G4_TISSUE_SOFT_ICRP"],
        [300, 800, "G4_B-100_BONE"],
        [800, 6000, "G4_BONE_COMPACT_ICRU"],
    ]
    # or alternatively, from a file (like in Gate)
    vm = read_voxel_materials(paths.gate_data / "patient-HU2mat-v1.txt")
    vm[0][0] = -2000
    # Note: the voxel_materials are turned into a structured array when setting the user info
    # Therefore, we store the previous one, assign the new one, and only then compare them!
    assert patient.voxel_materials == vm
    patient.voxel_materials = vm
    # write the image of labels (None by default)
    patient.dump_label_image = paths.output / "test009_label.mhd"

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 130 * MeV
    source.particle = "proton"
    source.position.type = "sphere"
    source.position.radius = 10 * mm
    source.position.translation = [0, 0, -14 * cm]
    source.activity = 10 * Bq
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]

    # cuts
    patient.set_production_cut(
        particle_name="electron",
        value=3 * mm,
    )

    # add dose actor
    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    dose_actor.attached_to = "patient"
    dose_actor.edep_uncertainty.active = True
    dose_actor.size = [99, 99, 99]
    dose_actor.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose_actor.output_coordinate_system = "attached_to_image"
    dose_actor.translation = [2 * mm, 3 * mm, -2 * mm]
    dose_actor.hit_type = "random"

    # add stat actor
    stats_actor = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats_actor.track_types_flag = True

    # print info
    print(sim.volume_manager.dump_volumes())

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run()

    # print results at the end
    print(stats_actor)
    print(dose_actor)

    utility.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.userhooks import check_production_cuts
from test013_phys_lists_helpers import (
    create_pl_sim,
)

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test013_phys_lists")

    # create simulation
    sim = create_pl_sim()

    # keep only ion sources
    sim.source_manager.sources.pop("gamma")

    # change physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = True
    # cuts = p.production_cuts
    mm = gate.g4_units.mm
    sim.physics_manager.global_production_cuts.gamma = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 0.1 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm
    sim.physics_manager.global_production_cuts.proton = 1 * mm

    reg = sim.physics_manager.add_region("reg")
    reg.production_cuts.gamma = 1 * mm
    reg.production_cuts.electron = 0.01 * mm
    reg.production_cuts.positron = 1 * mm
    reg.production_cuts.proton = 1 * mm
    reg.associate_volume("b1")

    # em parameters
    # phys_em_parameters(p)

    print("Phys list cuts:")
    print(sim.physics_manager.dump_production_cuts())
    print("Volume tree:")
    print(sim.volume_manager.dump_volume_tree())

    # start simulation
    sim.g4_verbose = False
    # sim.g4_commands_after_init.append("/tracking/verbose 1")
    sim.user_hook_after_init = check_production_cuts
    sim.run()

    # Gate mac/main_3.mac
    stats = sim.get_actor("Stats")
    stats_ref = utility.read_stat_file(paths.gate_output / "stat_3.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.1)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import opengate_core as g4

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test017")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.check_volumes_overlap = True
    sim.random_seed = 254123
    sim.output_dir = paths.output

    #  change world size
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    gcm3 = gate.g4_units.g_cm3

    sim.world.size = [1.5 * m, 1.5 * m, 1.5 * m]

    # add a simple volume
    airBox = sim.add_volume("Box", "AirBox")
    airBox.size = [30 * cm, 30 * cm, 30 * cm]
    airBox.translation = [0 * cm, 0 * cm, 0 * cm]
    airBox.material = "G4_AIR"
    airBox.color = [0, 0, 1, 1]  # blue

    # lyso material
    n = g4.G4NistManager.Instance()
    print(n)
    elems = ["Lu"]  # , 'Yttrium', 'Silicon', 'Oxygen']
    nbAtoms = [18]  # , 2, 10, 50]
    n.ConstructNewMaterialNbAtoms("LYSO", elems, nbAtoms, 7.1 * gcm3)

    # repeat a box
    crystal = sim.add_volume("Box", "crystal")
    crystal.mother = "AirBox"
    crystal.size = [1 * cm, 1 * cm, 1 * cm]
    # assign 4 translations -> this will create 4 physical copies of this volume in space
    crystal.translation = [
        [1 * cm, 0 * cm, 0],
        [0.2 * cm, 2 * cm, 0],
        [-0.2 * cm, 4 * cm, 0],
        [0, 6 * cm, 0],
    ]
    crystal.material = "LYSO"
    print(crystal)

    # WARNING:
    # For large number of repetition, look at test028 with RepeatParameterised volume
    # (it is more efficient)

    # default source for tests
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 0.511 * MeV
    source.position.type = "sphere"
    source.position.radius = 1 * cm
    source.position.translation = [0, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 1, 0]
    source.activity = 10000 * Bq

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.edep.output_filename = "test017.mhd"
    # dose.output = ref_path / 'test017-edep-ref.mhd'
    dose.attached_to = "crystal"
    dose.size = [150, 150, 150]
    dose.spacing = [1 * mm, 1 * mm, 1 * mm]
    dose.translation = [5 * mm, 0 * mm, 0 * mm]
    print(
        "The Dose actor is triggered every time a hit occurs in the (logical volume) "
        '"crystal" (and any of its associated repeated physical volumes).'
    )
    print(
        "The Dose actor is attached to the first (repeated) crystal, it moves with its coord system."
    )

    # start simulation
    sim.run()

    # print results
    print(stats)

    # tests
    stats_ref = utility.read_stat_file(paths.output_ref / "test017-stats-ref.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.04)
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test017-edep-ref.mhd",
            dose.edep.get_output_path(),
            stats,
            tolerance=70,
            sum_tolerance=6,
            ignore_value_data2=0,
            apply_ignore_mask_to_sum_check=False,  # force legacy behavior
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 30 15:54:30 2023

@author: fava
"""

import opengate as gate
import numpy as np
import re
from opengate.tests import utility


def read_table(path, skip_lines=4):
    table = dict()  # k: energy, v: dedx
    with open(path, "r") as f:
        lines = f.readlines()
    lines = lines[skip_lines:]
    lines = [
        re.findall(r"\d+\.\d+(?:[eE][+-]?\d+)?|\d+(?:[eE][+-]?\d+)?", l) for l in lines
    ]
    for l in lines:
        table[float(l[0])] = float(l[1])

    return table


paths = utility.get_default_test_paths(__file__, "test088_emcalc_actor", "test088")


sim = gate.Simulation()
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_engine = "MersenneTwister"
ui.random_seed = "auto"

m = gate.g4_units.m
cm = gate.g4_units.cm
km = gate.g4_units.km

world = sim.world
world.size = [3 * m, 3 * m, 3 * m]
world.material = "G4_AIR"

waterbox = sim.add_volume("Box", "Waterbox")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
waterbox.material = "G4_WATER"

sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"
sim.physics_manager.set_production_cut("world", "all", 1000 * km)

# EmCalc actor
em_calc = sim.add_actor("EmCalculatorActor", "test")
em_calc.attached_to = waterbox.name
em_calc.is_ion = True
# em_calc.particle_name = 'GenericIon'
em_calc.ion_params = "1 1"
em_calc.material = "G4_WATER"
# em_calc.nominal_energies = list(np.logspace(np.log10(1e-3), np.log10(1e3), 1000))
em_calc.nominal_energies = [10, 1e2, 1e3]
em_calc.savefile_path = paths.output / "dedx_table_H.txt"

stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

sim.run()

# compare against PSTAR table for protons in liquid water for some energies
pstar_tab_path = paths.output_ref / "PSTAR_table_water.txt"

ref_table = read_table(pstar_tab_path, skip_lines=5)
actor_table = read_table(paths.output / "dedx_table_H.txt", skip_lines=4)

is_ok = True
delta_thresh = 0.05

for e, dedx_actor in actor_table.items():
    if e in ref_table:
        dedx_ref = ref_table[e]
        delta_dedx_rel = abs(dedx_actor - dedx_ref) / dedx_ref
        is_ok = delta_dedx_rel <= delta_thresh and is_ok

utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests.utility import (
    get_default_test_paths,
    test_ok,
    print_test,
)
from opengate.utility import g4_units
from opengate.managers import Simulation

if __name__ == "__main__":
    paths = get_default_test_paths(__file__, "gate_test004_simulation_stats_actor")

    # sim
    sim = Simulation()
    sim.progress_bar = True
    sim.number_of_threads = 2
    sim.random_seed = 321654
    #   sim.running_verbose_level = EVENT

    # units
    m = g4_units.m
    cm = g4_units.cm
    keV = g4_units.keV
    mm = g4_units.mm
    um = g4_units.um
    Bq = g4_units.Bq
    s = g4_units.s

    # world
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    # wb
    waterbox = sim.add_volume("Box", "Waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    # phys
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"

    # src
    source = sim.add_source("GenericSource", "s1")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    # source.n = 200000
    source.activity = 100000 * Bq / sim.number_of_threads

    # src
    source = sim.add_source("GenericSource", "s2")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 100000 / sim.number_of_threads

    # src
    source = sim.add_source("GenericSource", "s3")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 100000 * Bq / sim.number_of_threads
    source.half_life = 0.5 * s

    # stats
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # go
    sim.run_timing_intervals = [
        [0, 1 * s],
        [1.5 * s, 3.0 * s],
        # Watch out : there is (on purpose) a 'hole' in the timeline
        [3.5 * s, 5.5 * s],
    ]
    sim.run(start_new_process=True)

    stats = sim.get_actor("Stats")
    print(stats)

    # Comparison with gate simulation
    n1 = sim.expected_number_of_events
    n2 = stats.counts.events
    f = abs(n1 - n2) / n2
    is_ok = f < 0.01
    print()
    print_test(is_ok, f"Predicted nb of events = {n1}, real = {n2}: {f*100:.2f}%")

    test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test053_phid_helpers2 import *
import opengate as gate


def main(dependency="test053_phid_10_all_ref_mt.py"):
    paths = get_default_test_paths(__file__, "", output_folder="test053")

    # bi213 83 213
    # ac225 89 225
    # fr221 87 221
    # pb 82 212
    # po 84 213
    # tl 81 209
    z = 89
    a = 225
    nuclide, _ = get_nuclide_and_direct_progeny(z, a)
    print(nuclide)
    sim_name = f"{nuclide.nuclide}_12_model_mt"

    # this test need the test053_phid_10 before
    root_ref = paths.output_ref / f"test053_{nuclide.nuclide}_10_ref.root"
    if not os.path.exists(root_ref):
        # ignore on windows
        if os.name == "nt":
            test_ok(True)
            sys.exit(0)
        cmd = "python " + str(paths.current / dependency)
        r = os.system(cmd)

    sim = gate.Simulation()
    create_sim_test053(sim, sim_name)

    # sources
    sim.number_of_threads = 3
    activity_in_Bq = 500
    s = add_source_model(sim, z, a, activity_in_Bq)
    s.atomic_relaxation_flag = True
    s.isomeric_transition_flag = True
    s.verbose = False

    # go
    sec = g4_units.second
    min = g4_units.minute
    start_time = 15 * min
    end_time = start_time + 2 * min
    duration = end_time - start_time
    print(f"start time {start_time / sec}")
    print(f"end time {end_time / sec}")
    print(f"Duration {duration / sec}")
    print(f"Ions {activity_in_Bq * duration / sec:.0f}")
    sim.run_timing_intervals = [[start_time, end_time]]

    # go
    sim.run(start_new_process=True)

    # print stats
    stats = sim.get_actor("stats")
    print(stats)

    # compare
    warning(f"check root files")
    root_ref = paths.output_ref / f"test053_{nuclide.nuclide}_10_ref.root"
    root_model = sim.get_actor("phsp").get_output_path()
    is_ok = compare_root_energy(
        root_ref,
        root_model,
        start_time,
        end_time,
        model_index=-1,
        tol=0.055,
        erange=[20, 600],
        n_tol=0.35,
    )

    test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import opengate as gate
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test032")

    # create the simulation
    sim = gate.Simulation()
    sim.output_dir = paths.output

    # shhhht !
    sim.verbose_level = gate.logger.NONE

    m = gate.g4_units.m
    cm = gate.g4_units.cm

    # world
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a iec phantom
    iec = gate_iec.add_iec_phantom(sim)

    extra_sphere = sim.add_volume("Sphere", name="extra_sphere")
    extra_sphere.rmax = 2 * gate.g4_units.cm
    extra_sphere.translation = [20 * cm, 20 * cm, 20 * cm]

    print("Automatic voxelization of all the geometry")
    # voxelize the geometry with 3x3x3 mm spacing
    labels_auto, image_auto = sim.voxelize_geometry(filename="test032_auto")
    info = gate.image.get_info_from_image(image_auto)
    print(f"Image (IEC and sphere): {info.size} {info.spacing} {info.origin}")

    print("Voxelization 3 mm")
    # voxelize the geometry with 3x3x3 mm spacing
    labels_3mm, image_3mm = sim.voxelize_geometry(
        iec, spacing=(3, 3, 3), margin=1, filename="test032_3mm"
    )
    info = gate.image.get_info_from_image(image_3mm)
    print(f"Image (3 mm spacing): {info.size} {info.spacing} {info.origin}")

    print("Voxelization 1 mm")
    # voxelize the geometry with 1x1x1 mm spacing
    labels_1mm, image_1mm, path_to_image_1mm = sim.voxelize_geometry(
        iec, spacing=(1, 1, 1), margin=1, filename="test032_1mm", return_path=True
    )
    info = gate.image.get_info_from_image(image_1mm)
    print(f"Image (1 mm spacing): {info.size} {info.spacing} {info.origin}")

    # read and compare labels
    gate.exception.warning("\nDifference labels")
    with open(paths.output_ref / "test032_labels.json", "r") as f:
        ref_labels = json.load(f)
    added, removed, modified, same = utility.dict_compare(ref_labels, labels_1mm)
    is_ok = len(added) == 0 and len(removed) == 0 and len(modified) == 0
    utility.print_test(is_ok, f"Labels comparisons, added:    {added}")
    utility.print_test(is_ok, f"Labels comparisons, removed:  {removed}")
    utility.print_test(is_ok, f"Labels comparisons: modified: {modified}")

    # compare images
    gate.exception.warning("\nDifference with ref image")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test032_iec.mhd",
            path_to_image_1mm,
            stats=None,
            tolerance=0.01,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test019_linac_phsp_helpers as test019

if __name__ == "__main__":
    sim = test019.init_test019(1, version="linacphsp")

    s = sim.volume_manager.print_volume_tree()
    source = sim.get_source_user_info("Default")

    test019.run_test019(sim)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test052")

    """
    Test the option in GenericSource to use a TAC Time Activity Curve
    """

    # create the simulation
    sim = gate.Simulation()

    # units
    m = gate.g4_units.m
    nm = gate.g4_units.nm
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    Bq = gate.g4_units.Bq
    kBq = 1e3 * Bq
    sec = gate.g4_units.s
    min = gate.g4_units.min
    keV = gate.g4_units.keV

    # verbose
    sim.visu = False
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.number_of_threads = 1
    sim.random_seed = 987654321
    sim.output_dir = paths.output

    # world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_AIR"

    # source of Ac225 ion
    s1 = sim.add_source("GenericSource", "source")
    s1.particle = "gamma"
    s1.position.type = "sphere"
    s1.position.radius = 1 * nm
    s1.position.translation = [0, 0, 0]
    s1.direction.type = "iso"
    s1.energy.mono = 140 * keV

    # create a decay TAC bw 2 and 5 sec
    starting_activity = 100 * kBq / sim.number_of_threads
    half_life = 2 * sec
    times = np.linspace(1, 6, num=500, endpoint=True) * sec
    decay = np.log(2) / half_life
    activities = [starting_activity * np.exp(-decay * t) for t in times]
    s1.tac_times = times
    s1.tac_activities = activities

    # stats
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # phsp actor for timing
    phsp = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp.attributes = ["GlobalTime"]
    phsp.steps_to_store = "exiting"
    phsp.output_filename = "test052_tac.root"

    # go
    # sim.running_verbose_level = gate.EVENT
    # (purposely a "hole" in the runtime ; warning not too large otherwise the fit fails)
    sim.run_timing_intervals = [[0, 2.9 * sec], [3 * sec, 7 * sec]]
    sim.run(start_new_process=True)

    # print
    print(stats)

    # check root
    print()
    gate.exception.warning("Check root time")
    root1, n1 = utility.open_root_as_np(phsp.get_output_path(), "phsp")
    etimes = root1["GlobalTime"] / sec
    print(f"Number of events : {len(etimes)}")

    # check fit (it is important to convert 'etimes' to sec otherwise
    # the fit fails)
    hl, xx, yy = utility.fit_exponential_decay(
        etimes, 0, 7
    )  # times[0] / sec, times[-1] / sec)
    tol = 0.05
    hl_ref = half_life / sec
    diff = abs(hl - hl_ref) / hl_ref
    is_ok = b = diff < tol
    diff *= 100
    utility.print_test(b, f"Half life {hl_ref:.2f} sec vs {hl:.2f} sec : {diff:.2f}% ")

    # plot the fit
    f, ax = plt.subplots(1, 1, figsize=(25, 10))
    utility.plot_hist(ax, etimes, f"Events times")
    ax.plot(xx, yy, label=f"fit")
    ax.legend()

    fn = paths.output / "test052_tac_times.png"
    plt.savefig(fn)
    print(f"Plot in {fn}")

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test053_phid_helpers1 import *
import matplotlib.pyplot as plt
import radioactivedecay as rd
import opengate as gate

if __name__ == "__main__":
    paths = get_default_test_paths(__file__, "", output_folder="test053")

    # ac225
    z = 89
    a = 225

    # bi213
    # z = 83
    # a = 213

    """
    In a water world, simulate an ion source. The run time is from 0 to 1 hour.
    The source activity
    Store a phsp with only the produced ions (filter gamma, nu, alpha, e-).
    """

    # create simulation
    sim = gate.Simulation()
    sim.progress_bar = True
    ion_name, _ = create_ion_gamma_simulation(sim, paths, z, a)

    # get list of nuclide organized per ion
    nuclide = get_nuclide_from_name(ion_name)
    decay_list = get_nuclide_progeny(nuclide)
    decay_list_per_ion = {}
    for d in decay_list:
        decay_list_per_ion[d.nuclide.nuclide.replace("-", "")] = d
    for d in decay_list_per_ion:
        print(d, decay_list_per_ion[d])

    # change simulation parameters
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.s
    h = gate.g4_units.h
    activity = 10 * Bq
    end = 1 * h
    update_sim_for_tac(sim, ion_name, nuclide, activity, end)

    # --------------------------------------------------------------------------
    # go
    # sim.running_verbose_level = gate.EVENT
    sim.run()
    end = end / sec
    # --------------------------------------------------------------------------

    # print
    stats = sim.get_actor("stats")
    print(stats)

    # analyse
    time_by_ion_final = analyse_time_per_ion_root(sim, end)

    # plot
    lines_colour_cycle = [p["color"] for p in plt.rcParams["axes.prop_cycle"]]
    colors = {}
    i = 0
    rad_list = []
    for d in decay_list:
        if d.hl == float("inf"):
            continue
        rad_list.append(d.nuclide.nuclide)
        n = d.nuclide.nuclide.replace("-", "")
        colors[n] = lines_colour_cycle[i]
        i += 1

    fig, ax = plt.subplots(1, 1, figsize=(20, 10))
    inv = rd.Inventory({nuclide.nuclide: activity / Bq}, "Bq")

    inv.plot(
        end,
        "s",
        yunits="Bq",
        fig=fig,
        axes=ax,
        alpha=0.2,
        linewidth=8,
        order="dataset",
        display=rad_list,
    )

    i = 0
    bins = 200

    # random check ?
    check = [0.10, 0.20, 0.50, 0.7]
    check_dc = []
    for c in check:
        dc = inv.decay(end * c, "s")
        check_dc.append(dc)

    is_ok = True
    for ion_name in decay_list_per_ion:
        # print(f'ion {ion_name}')
        b = decay_list_per_ion[ion_name]
        daughters = b.nuclide.progeny()
        branching_fractions = nuclide.branching_fractions()
        x = np.array([])
        for d, br in zip(daughters, branching_fractions):
            d_name = d.replace("-", "")
            a = np.array(time_by_ion_final[d_name])
            # print(f'\t {d_name} {b.intensity} x={len(x)} a={len(a)}')
            x = np.concatenate((x, a))

        if len(x) == 0:
            continue
        col = colors[ion_name]
        w = np.ones_like(x) * b.intensity
        hist, bin_edges = np.histogram(x, bins=bins, range=[0, end], weights=w)
        dx = bin_edges[1] - bin_edges[0]
        hist = hist / dx
        ax.stairs(hist, bin_edges, label=f"{ion_name}", color=col)

        # check
        i = 0
        n = b.nuclide.nuclide
        tol = 30
        for c in check:
            index = int(bins * c)
            ac = check_dc[i].activities("Bq")[n]
            if ac < 1:
                tol = 700
            diff = np.fabs(ac - hist[index]) / ac * 100.0
            ok = diff < tol
            print_test(
                ok,
                f"check {ion_name} time={end * c:.2f}   ref = {ac:.2f} vs {hist[index]:.2f} ->  {diff:.2f}  (tol = {tol:.2f}%)",
            )
            is_ok = ok and is_ok
            i += 1

    ax.legend()
    # plt.show()

    f = paths.output / "test053_gamma_from_ion_decay_tac_ac225.png"
    print(f"Plot save in {f}")
    plt.savefig(f)

    test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.contrib.spect.pytomography_helpers import *
from opengate.tests import utility
import subprocess


def main(dependency="test094b_spect_build_pytomography.py"):

    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test094_pytomography"
    )

    # The test needs the output of the other tests
    if not os.path.isfile(paths.output / "sinogram.raw"):
        subprocess.call(["python", paths.current / dependency])

    # read the metadata file
    json_file = paths.output / "pytomography_gate.json"
    metadata = pytomography_read_metadata(json_file)

    # reconstruction
    keV = g4_units.keV
    img = pytomography_osem_reconstruction(
        metadata,
        index_peak=4,
        psf_photon_energy=208 * keV,
        attenuation_flag=True,
        scatter_mode="TEW",
        n_iters=4,
        n_subsets=6,
    )

    # write the image
    output = paths.output / "reconstructed.mhd"
    print(f"Save reconstructed image in {output}")
    sitk.WriteImage(img, output)

    # compare the image with the reference reconstructed one
    ref_output = paths.output_ref / "reconstructed.mhd"
    ref = sitk.ReadImage(ref_output)

    diff = sitk.GetArrayFromImage(sitk.Abs(ref - img))
    print(f"Max diff: {np.max(diff)}")
    # assert np.max(diff) < 1e-3, "Error in the reconstructed image"

    is_ok = utility.assert_images(
        ref_output,
        output,
        stats=None,
        tolerance=0.1,
        fig_name=paths.output / "094.png",
        sad_profile_tolerance=0.1,
    )

    print("Check with :")
    print(
        "vv ../data/ct_5mm.mhd --fusion ../data/3_sources_5mm_v2.mhd ../data/ct_5mm.mhd --fusion ../output/test094_pytomography/reconstructed.mhd --linkall"
    )

    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



import opengate as gate
import uproot
from opengate.exception import GateImplementationError
import opengate.tests.utility as utility


g_cm3 = gate.g4_units.g_cm3
mm = gate.g4_units.mm
Bq = gate.g4_units.Bq
keV = gate.g4_units.keV
sec = gate.g4_units.second
ns = gate.g4_units.nanosecond
deg = gate.g4_units.degree


def create_materials(sim):
    sim.volume_manager.material_database.add_material_nb_atoms(
        "LYSO", ["Lu", "Y", "Si", "O"], [18, 2, 10, 50], 7.1 * g_cm3
    )


if __name__ == "__main__":

    paths = utility.get_default_test_paths(__file__, output_folder="test086")

    sim = gate.Simulation()
    sim.g4_verbose = False
    sim.visu = False
    sim.visu_type = "vrml"
    sim.random_engine = "MersenneTwister"
    sim.random_seed = 1234
    sim.output_dir = paths.output
    sim.number_of_threads = 1
    create_materials(sim)

    sim.world.size = [200 * mm, 200 * mm, 200 * mm]
    sim.world.material = "G4_AIR"

    block_size = 50 * mm

    # Two blocks of LYSO, created separately without repetition, symmetrically along the x-axis.
    crystal_no_repeat1 = sim.add_volume("Box", "crystal_no_repeat1")
    crystal_no_repeat1.size = [block_size] * 3
    crystal_no_repeat1.translation = [-block_size * 1.1, 0, 0]
    crystal_no_repeat1.material = "LYSO"
    crystal_no_repeat2 = sim.add_volume("Box", "crystal_no_repeat2")
    crystal_no_repeat2.size = [block_size] * 3
    crystal_no_repeat2.translation = [block_size * 1.1, 0, 0]
    crystal_no_repeat2.material = "LYSO"

    # Two blocks of LYSO, created with repetition, symmetrically along the y-axis.
    crystal_repeat = sim.add_volume("Box", "crystal_repeat")
    crystal_repeat.size = [block_size] * 3
    crystal_repeat.translation = gate.geometry.utility.get_grid_repetition(
        [1, 2, 1], [0, 2 * block_size * 1.1, 0]
    )
    crystal_repeat.material = "LYSO"

    # Gamma 511 keV source, radiating in the x-y plane.
    source = sim.add_source("GenericSource", "source")
    source.particle = "gamma"
    source.activity = 100 * Bq
    source.energy.type = "mono"
    source.energy.mono = 511 * keV
    source.position.type = "sphere"
    source.position.radius = 0.5 * mm
    source.direction.type = "iso"
    source.direction.theta = [90 * deg, 90 * deg]

    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"

    stats = sim.add_actor("SimulationStatisticsActor", "stats")

    hc = sim.add_actor("DigitizerHitsCollectionActor", "hits")
    hc.attached_to = [crystal_no_repeat1.name, crystal_no_repeat2, crystal_repeat]
    hc.output_filename = "sim.root"
    hc.attributes = [
        "EventID",
        "PreStepUniqueVolumeID",
        "TrackVolumeName",
        "PostPosition",
        "TotalEnergyDeposit",
        "GlobalTime",
    ]

    sc = sim.add_actor("DigitizerAdderActor", "singles")
    sc.attached_to = hc.attached_to
    sc.input_digi_collection = hc.name
    sc.policy = "EnergyWinnerPosition"
    sc.output_filename = hc.output_filename

    sim.run_timing_intervals = [[0, 1 * sec]]

    sim.run()

    root_file = uproot.open(paths.output / hc.output_filename)
    singles_tree = root_file["singles"]

    unique_volume_ids = set(singles_tree["PreStepUniqueVolumeID"].array())
    num_unique_hit_volumes = len(unique_volume_ids)
    print(f"{num_unique_hit_volumes} volumes have been hit: {unique_volume_ids}")

    is_ok = True
    exceptions = []

    expected_num_volumes = 4
    if num_unique_hit_volumes < expected_num_volumes:
        is_ok = False
        exceptions.append(
            GateImplementationError(
                f"Hits have been recorded for only {num_unique_hit_volumes} out of {expected_num_volumes} volumes: {unique_volume_ids}"
            )
        )
    utility.test_ok(is_ok, exceptions=exceptions)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.utility import g4_units
import opengate.tests.utility as utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test004_simulation_stats_actor"
    )
    sim = gate.Simulation()

    m = g4_units.m
    cm = g4_units.cm
    keV = g4_units.keV
    mm = g4_units.mm
    um = g4_units.um
    Bq = g4_units.Bq

    waterbox = sim.add_volume("Box", "Waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 200000

    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # run
    sim.run(start_new_process=True)
    print(stats)

    # Comparison with gate simulation
    # gate_test4_simulation_stats_actor
    # Gate mac/main.mac
    stats_ref = utility.read_stat_file(paths.gate_output / "stat.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.01)

    utility.test_ok(is_ok)



import json
import itk
import numpy as np
import math

from opengate import utility
import opengate.geometry.volumes
from opengate.utility import fatal, g4_units
from opengate.geometry.volumes import unite_volumes
from opengate.sources.gansources import generate_isotropic_directions

iec_plastic = "IEC_PLASTIC"
water = "G4_WATER"
iec_lung = "G4_LUNG_ICRP"

# colors
red = [1, 0.7, 0.7, 0.8]
blue = [0.5, 0.5, 1, 0.8]
gray = [0.5, 0.5, 0.5, 1]
transparent = [0, 0, 0, 0]


def create_material(simulation):
    elems = ["C", "H", "O"]
    nb_atoms = [5, 8, 2]
    gcm3 = g4_units.g_cm3
    simulation.volume_manager.material_database.add_material_nb_atoms(
        "IEC_PLASTIC", elems, nb_atoms, 1.18 * gcm3
    )


def add_iec_phantom(
    simulation,
    name="iec",
    check_overlap=False,
    sphere_starting_angle=False,
    toggle_sphere_order=False,
):
    # https://www.nuclemed.be/product.php?cat=102&prod=297 ???
    # unit
    mm = g4_units.mm
    create_material(simulation)

    # check overlap only for debug
    original_check_overlap_flag = simulation.check_volumes_overlap
    simulation.check_volumes_overlap = check_overlap

    # Outside structure
    iec, _, _ = add_iec_body(simulation, name)
    iec.material = iec_plastic
    iec.color = red

    # Inside space for the water, same as the shell, with 3 mm less
    thickness = 3 * mm
    thickness_z = 10 * mm
    interior, top_interior, _ = add_iec_body(
        simulation, f"{name}_interior", thickness, thickness_z
    )
    interior.mother = iec.name
    interior.material = water
    interior.color = blue

    # central tube in iec_plastic
    add_iec_central_cylinder(simulation, name, top_interior)

    # spheres
    add_iec_all_spheres(
        simulation, name, thickness_z, sphere_starting_angle, toggle_sphere_order
    )

    simulation.check_volumes_overlap = original_check_overlap_flag
    return iec


def add_iec_body(simulation, name, thickness=0.0, thickness_z=0.0):
    cm = g4_units.cm
    nm = g4_units.nm
    deg = g4_units.deg

    # total length
    length = 21.4 * cm

    # top
    top_shell = opengate.geometry.volumes.TubsVolume(name=f"{name}_top_shell")
    top_shell.rmax = 15 * cm - thickness
    top_shell.rmin = 0
    top_shell.dz = length / 2 - thickness_z
    top_shell.sphi = 0 * deg
    top_shell.dphi = 180 * deg

    # Lower left half of phantom
    bottom_left_shell = opengate.geometry.volumes.TubsVolume(
        name=f"{name}_bottom_left_shell"
    )
    bottom_left_shell.rmax = 8 * cm - thickness
    bottom_left_shell.rmin = 0
    bottom_left_shell.dz = length / 2 - thickness_z
    bottom_left_shell.sphi = 270 * deg
    bottom_left_shell.dphi = 90 * deg

    # Lower right half of phantom
    bottom_right_shell = opengate.geometry.volumes.TubsVolume(
        name=f"{name}_bottom_right_shell"
    )
    bottom_right_shell.configure_like(bottom_left_shell)
    bottom_right_shell.sphi = 180 * deg
    bottom_right_shell.dphi = 90 * deg

    # slightly move the volumes to avoid large surface overlap during the union
    # unsure if it worth it
    tiny = 1 * nm

    # Bottom box
    # length = Z = sup-inf = is 21.4
    # bottom radius = Y = ant-post = 8 cm
    # width = X = left-right = in between the two bottom rounded  = 14 * cm
    # X total is 14 + 8 + 8 = 30 cm (main radius is 15cm)
    bottom_central_shell = opengate.geometry.volumes.BoxVolume(
        name=f"{name}_bottom_central_shell"
    )
    bottom_central_shell.size = [14 * cm + tiny, 8 * cm, length]
    bottom_central_shell.size[1] -= thickness
    bottom_central_shell.size[2] -= 2 * thickness_z
    c = -bottom_central_shell.size[1] / 2 + tiny

    # union
    t_bc = unite_volumes(top_shell, bottom_central_shell, translation=[0, c, 0])
    t_bc_bl = unite_volumes(
        t_bc, bottom_left_shell, translation=[7 * cm - tiny, tiny, 0]
    )
    iec = unite_volumes(
        t_bc_bl,
        bottom_right_shell,
        translation=[-7 * cm + tiny, tiny, 0],
        new_name=name,
    )
    simulation.volume_manager.add_volume(iec)

    return iec, top_shell, c


def add_iec_central_cylinder(sim, name, top_interior):
    # unit
    cm = g4_units.cm
    deg = g4_units.deg

    cc = sim.add_volume("Tubs", f"{name}_center_cylinder")
    cc.mother = f"{name}_interior"
    cc.rmax = 2.5 * cm
    cc.rmin = 2.1 * cm
    cc.dz = top_interior.dz
    cc.sphi = 0 * deg
    cc.dphi = 360 * deg
    cc.material = iec_plastic
    cc.translation = [0, 3.5 * cm, 0]
    cc.color = red

    # central tube lung material
    hscc = sim.add_volume("Tubs", f"{name}_center_cylinder_hole")
    hscc.mother = f"{name}_interior"
    hscc.rmax = 2.1 * cm
    hscc.rmin = 0 * cm
    hscc.dz = top_interior.dz
    hscc.material = iec_lung
    hscc.translation = [0, 3.5 * cm, 0]
    hscc.color = gray


def add_iec_all_spheres(
    simulation, name, thickness_z, starting_angle=False, reverse_order=False
):
    """
    Starting angle : in deg. Indicate the (angle) position of the first smallest sphere.
    It is 180 deg by default.
    """
    # unit
    cm = g4_units.cm
    mm = g4_units.mm
    deg = g4_units.deg

    """
    The spheres are positioned in a circle every 60 deg.
    It can be modified.
    """
    v = f"{name}_interior"
    h_relative = 2.7 * cm
    r = 11.45367 * cm / 2
    ang = 360 / 6 * deg
    if starting_angle is False:
        starting_angle = 3 * ang
    a = starting_angle

    spheres_diam = [37, 28, 22, 17, 13, 10]
    if reverse_order:
        spheres_diam.reverse()

    for sd in spheres_diam:
        px = np.cos(a) * r
        py = np.sin(a) * r + 3.5 * cm
        add_iec_one_sphere(
            simulation,
            name,
            v,
            sd * mm,
            1 * mm,
            3.5 * mm,
            [px, py, h_relative],
            thickness_z,
        )
        a += ang


def add_iec_one_sphere(
    sim, name, vol, diam, sph_thick, cap_thick, position, thickness_z
):
    mm = g4_units.mm
    cm = g4_units.cm
    d = f"{(diam / mm):.0f}mm"
    rad = diam / 2
    h_relative = position[2]

    # interior sphere
    sph = sim.add_volume("Sphere", f"{name}_sphere_{d}")
    sph.mother = vol
    sph.translation = np.array(position)  # need to copy the array!
    sph.rmax = rad
    sph.rmin = 0
    sph.material = "G4_WATER"

    # outer sphere shell
    sphs = sim.add_volume("Sphere", f"{name}_sphere_shell_{d}")
    sphs.mother = vol
    sphs.translation = np.array(position)
    sphs.rmax = rad + sph_thick
    sphs.rmin = rad
    sphs.material = iec_plastic

    # capillary
    cap = sim.add_volume("Tubs", f"{name}_capillary_{d}")
    cap.mother = vol
    cap.translation = np.array(position)
    cap.material = "G4_WATER"
    cap.rmax = 0.25 * cm
    cap.rmin = 0 * cm
    # 21.4/2 = 10.7 interior height (top_interior)
    h = 21.4 / 2 * cm - thickness_z
    cap.dz = (h - h_relative - rad - sph_thick) / 2.0
    cap.translation[2] = h_relative + rad + sph_thick + cap.dz

    # capillary outer shell
    caps = sim.add_volume("Tubs", f"{name}_capillary_shell_{d}")
    caps.configure_like(cap)
    caps.material = iec_plastic
    caps.rmax = cap_thick
    caps.rmin = cap.rmax


def add_spheres_sources(
    simulation,
    iec_name,
    src_name,
    spheres,
    activity_Bq_mL,
    verbose=False,
    source_type="GenericSource",
):
    spheres_diam = [10, 13, 17, 22, 28, 37]
    sources = []
    if spheres == "all":
        spheres = spheres_diam
    for sphere, ac in zip(spheres, activity_Bq_mL):
        if sphere in spheres_diam:
            if ac > 0:
                s = add_one_sphere_source(
                    simulation,
                    iec_name,
                    src_name,
                    float(sphere),
                    float(ac),
                    source_type=source_type,
                )
                sources.append(s)
        else:
            fatal(
                f"Error the sphere of diameter {sphere} does not exists in {spheres_diam}"
            )
    # verbose ?
    if verbose:
        s = dump_spheres_activity(simulation, iec_name, src_name)
        t = compute_total_spheres_activity(simulation, iec_name, src_name)
        print(s)
        print(f"Total activity is {t} Bq")
    return sources


def add_spheres_sources_equal(sim, iec_name, src_name, total_activity):
    Bq = g4_units.Bq
    sources = add_spheres_sources(sim, iec_name, src_name, "all", [1.0] * 6)
    t = compute_total_spheres_activity(sim, iec_name, src_name) * Bq
    for source in sources:
        # set the total activity to the asked number of particle
        source.activity = (source.activity / t) * total_activity
    return sources


def compute_sphere_activity(simulation, iec_name, src_name, diam):
    mm = g4_units.mm
    cm3 = g4_units.cm3
    Bq = g4_units.Bq
    d = f"{(diam / mm):.0f}mm"
    sname = f"{src_name}_{iec_name}_{d}"
    if sname not in simulation.source_manager.sources:
        return None, None, None, None
    src = simulation.source_manager.get_source(sname)
    vname = src.attached_to
    v = simulation.volume_manager.volumes[vname]
    s = v.solid_info
    ac = src.activity
    return ac / Bq, s.cubic_volume / cm3, sname, vname


def compute_total_spheres_activity(simulation, iec_name, src_name):
    spheres_diam = [10, 13, 17, 22, 28, 37]
    a = 0
    for diam in spheres_diam:
        ac, _, _, _ = compute_sphere_activity(simulation, iec_name, src_name, diam)
        if ac is None:
            continue
        a += ac
    return a


def dump_spheres_activity(simulation, iec_name, src_name):
    spheres_diam = [10, 13, 17, 22, 28, 37]
    out = ""
    i = 0
    for diam in spheres_diam:
        ac, vol, sname, vname = compute_sphere_activity(
            simulation, iec_name, src_name, diam
        )
        if ac is None:
            continue
        out += (
            f"{vname:<20} {sname:<20} "
            f"{vol:10.2f} mL   {ac:10.2f} Bq   {ac / vol:10.2f} Bq/mL\n"
        )
        i += 1
    return out[:-1]


def dump_bg_activity(simulation, iec_name, src_name):
    cm3 = g4_units.cm3
    Bq = g4_units.Bq
    BqmL = Bq / cm3
    sname = f"{iec_name}_{src_name}"
    if sname not in simulation.source_manager.sources:
        return
    src = simulation.source_manager.get_source(sname)
    v = simulation.volume_manager.volumes[src.attached_to]
    s = v.solid_info
    ac = src.activity
    out = (
        f"Volume = {v.name:<20} Source = {sname:<20} "
        f"{s.cubic_volume / cm3:10.2f} mL   {ac / Bq:10.2f} Bq   {ac / s.cubic_volume / BqmL:10.2f} Bq/mL"
    )
    return out


def add_one_sphere_source(
    simulation, iec_name, src_name, diameter, activity_Bq_mL, source_type
):
    mm = g4_units.mm
    mL = g4_units.mL
    d = f"{(diameter / mm):.0f}mm"
    sname = f"{iec_name}_sphere_{d}"

    # compute volume in mL (and check)
    volume_ref = 4 / 3 * np.pi * np.power(diameter / mm / 2, 3) * 0.001
    v = simulation.volume_manager.volumes[sname]
    s = v.solid_info
    volume = s.cubic_volume / mL
    if not math.isclose(volume_ref, volume, rel_tol=1e-7):
        fatal(
            f"Error while estimating the sphere volume {sname}: {volume_ref} vs {volume}"
        )

    source = simulation.add_source(source_type, f"{src_name}_{iec_name}_{d}")
    source.attached_to = sname
    # default values
    source.particle = "e+"
    source.energy.type = "F18"
    source.direction.type = "iso"
    source.activity = activity_Bq_mL * s.cubic_volume
    source.position.type = "sphere"
    source.position.radius = diameter / 2 * mm
    source.position.translation = [0, 0, 0]
    return source


def add_central_cylinder_source(
    simulation, iec_name, src_name, activity_Bq_mL, verbose=False
):
    # source
    bg = simulation.add_source("GenericSource", f"{iec_name}_{src_name}")
    bg.attached_to = f"{iec_name}_center_cylinder_hole"
    v = simulation.volume_manager.volumes[bg.attached_to]
    s = v.solid_info
    # (1 cm3 = 1 mL)
    bg.position.type = "box"
    bg.position.size = simulation.volume_manager.volumes[
        bg.attached_to
    ].bounding_box_size
    # this source is confined only within the mother volume, it does not include daughter volumes
    # it is a tubs inside the box
    bg.position.confine = bg.attached_to
    bg.particle = "e+"
    bg.energy.type = "F18"
    bg.activity = activity_Bq_mL * s.cubic_volume
    # verbose ?
    if verbose:
        # print(f"Bg volume {s.cubic_volume} cc")
        s = dump_bg_activity(simulation, iec_name, src_name)
        print(s)
    return bg


def add_background_source(
    simulation, iec_name, src_name, activity_Bq_mL, verbose=False
):
    # source
    bg = simulation.add_source("GenericSource", f"{iec_name}_{src_name}")
    bg.attached_to = f"{iec_name}_interior"
    v = simulation.volume_manager.volumes[bg.attached_to]
    s = v.solid_info
    # (1 cm3 = 1 mL)
    bg.position.type = "box"
    bg.position.size = simulation.volume_manager.volumes[
        bg.attached_to
    ].bounding_box_size
    # this source is confined only within the mother volume, it does not include daughter volumes
    bg.position.confine = bg.attached_to
    bg.particle = "e+"
    bg.energy.type = "F18"
    bg.activity = activity_Bq_mL * s.cubic_volume
    # the confine procedure from G4 seems to be confused when using a boolean solid like {iec_name}_interior
    # (or I did understand correctly how it works)
    # so, we need to move the source for correct sampling of the volume
    mm = g4_units.mm
    bg.position.translation = [0, 35 * mm, 0]
    # verbose ?
    if verbose:
        # print(f"Bg volume {s.cubic_volume} cc")
        s = dump_bg_activity(simulation, iec_name, src_name)
        print(s)
    return bg


def generate_pos_dir_one_sphere(center, radius, n, rs=np.random):
    """
    This function should be useful to generate conditional data for condGAN.
    It samples the position in a sphere and isotropic direction.
    The center/radius is the center and radius of the sphere
    A numpy array of (n,6) is returned.
    3 first = position
    3 last = direction
    """
    # uniform random vector of size n
    p = generate_pos_one_sphere(center, radius, n, rs)
    # direction
    v = generate_isotropic_directions(n, rs=rs)
    # concat all
    return np.column_stack((p, v))


def generate_pos_one_sphere(center, radius, n, rs=np.random):
    # uniform random vector of size n
    u = rs.uniform(0, 1, size=n)
    r = np.cbrt((u * radius**3))
    phi = rs.uniform(0, 2 * np.pi, n)
    theta = np.arccos(rs.uniform(-1, 1, n))
    # position in cartesian
    x = r * np.sin(theta) * np.cos(phi) + center[0]
    y = r * np.sin(theta) * np.sin(phi) + center[1]
    z = r * np.cos(theta) + center[2]
    # concat all
    return np.column_stack((x, y, z))


def generate_pos_dir_spheres(centers, radius, n_samples, shuffle, rs=np.random):
    """
    This function generate conditional data for condGAN.
    It samples the position in several spheres, with isotropic direction.
    The center/radius are the center and radius of the spheres.
    n_samples is the number of samples per sphere, with a total of n.
    Samples can be shuffled (by default).
    A numpy array of (n,6) is returned.
    """
    cond = None
    for rad, center, n in zip(radius, centers, n_samples):
        # approximate -> if the last one we complete to reach n
        x = generate_pos_dir_one_sphere(center, rad, n, rs=rs)
        if cond is None:
            cond = x
        else:
            cond = np.vstack((cond, x))

    # shuffle
    if shuffle:
        # it seems that permutation is **much** faster than shuffle
        # (checked 2022/06/047 on osx)

        # https://github.com/numpy/numpy/issues/11013
        # sstart = time.time()
        # np.random.shuffle(cond)
        # send = time.time()
        # print(f'shuffle 1 {send - sstart:0.4f} sec')

        # sstart = time.time()
        cond = cond.take(rs.permutation(cond.shape[0]), axis=0)
        # send = time.time()
        # print(f'shuffle 2 {send - sstart:0.4f} sec')

    return cond


def generate_pos_spheres(centers, radius, n_samples, shuffle, rs=np.random):
    """
    Like generate_pos_dir_spheres, but position only
    """
    cond = None
    for rad, center, n in zip(radius, centers, n_samples):
        # approximate -> if the last one we complete to reach n
        x = generate_pos_one_sphere(center, rad, n, rs)
        if cond is None:
            cond = x
        else:
            cond = np.vstack((cond, x))

    # shuffle
    if shuffle:
        cond = cond.take(rs.permutation(cond.shape[0]), axis=0)

    return cond


def get_n_samples_from_ratio(n, ratio):
    """
    For a given proportion of activities (in ratio) and total number of particle n,
    compute the list of particle for each index.
    """
    i = 0
    total = 0
    n_samples = []
    for r in ratio:
        if i == len(ratio) - 1:
            # last one ?
            m = n - total
        else:
            m = int(round(n * r))
        n_samples.append(m)
        total += m
        i += 1
    return n_samples


def compute_sphere_centers_and_volumes_OLD_NEVER_CALLED(sim, name):
    spheres_diam = [10, 13, 17, 22, 28, 37]
    centers = []
    volumes = []
    mm = g4_units.mm
    for diam in spheres_diam:
        # retrieve the name of the sphere volume
        d = f"{(diam / mm):.0f}mm"
        v = sim.volume_manager.volumes[f"{name}_sphere_{d}"]
        s = v.solid_info
        # from the solid get the center position
        center = v.translation
        centers.append(center)
        # and the volume
        volumes.append(s.cubic_volume)
    return centers, volumes


def get_default_sphere_centers_and_volumes_old():
    """
    Global spheres centers in the phantom, to avoid using the phantom in same cases.
    Were computed with 10/06/2022 version.
    No translation. To be recomputed with compute_sphere_centers_and_volumes
    """
    centers = [
        [28.6, -16.0367, 37.0],
        [-28.6, -16.0367, 37.0],
        [-57.2, 35.0, 37.0],
        [-28.6, 84.5367, 37.0],
        [28.6, 84.5367, 37.0],
        [57.2, 35.0, 37.0],
    ]
    volumes = [
        523.5987755982989,
        1150.3465099894627,
        2572.4407845144424,
        5575.279762570685,
        11494.040321933857,
        26521.84878038063,
    ]
    return centers, volumes


def get_default_sphere_centers_and_volumes():
    """
    Global spheres centers in the phantom, to avoid using the phantom in same cases.
    Were computed with 23/08/2023 version.
    No translation. To be recomputed with compute_sphere_centers_and_volumes
    """
    centers = [
        [-28.634175, 84.59584593, 27.0],
        [28.634175, 84.59584593, 27.0],
        [57.26835, 35.0, 27.0],
        [28.634175, -14.59584593, 27.0],
        [-28.634175, -14.59584593, 27.0],
        [-57.26835, 35.0, 27.0],
    ]
    volumes = [
        523.5987755982989,
        1150.3465099894627,
        2572.4407845144424,
        5575.279762570685,
        11494.040321933857,
        26521.84878038063,
    ]
    return centers, volumes


def add_iec_phantom_vox_FIXME_TO_REMOVE(sim, name, image_filename, labels_filename):
    iec = sim.add_volume("Image", name)
    iec.image = image_filename
    iec.material = "IEC_PLASTIC"
    labels = utility.read_json_file(labels_filename)
    iec.voxel_materials = []
    create_material(sim)
    material_list = {}
    for l in labels:
        mat = "IEC_PLASTIC"
        if "capillary" in l:
            mat = "G4_WATER"
        if "cylinder_hole" in l:
            mat = "G4_AIR"
        if "world" in l:
            mat = "G4_AIR"
        if "interior" in l:
            mat = "G4_WATER"
        if "sphere" in l:
            mat = "G4_WATER"
        if "shell" in l:
            mat = "IEC_PLASTIC"
        material_list[l] = mat
        m = [labels[l]["label"], labels[l]["label"] + 1, mat]
        iec.voxel_materials.append(m)
    return iec, material_list


def create_iec_phantom_source_vox(
    image_filename, labels_filename, source_filename, activities=None
):
    if activities is None:
        activities = {
            "iec_sphere_10mm": 1.0,
            "iec_sphere_13mm": 1.0,
            "iec_sphere_17mm": 1.0,
            "iec_sphere_22mm": 1.0,
            "iec_sphere_28mm": 1.0,
            "iec_sphere_37mm": 1.0,
        }

    img = itk.imread(image_filename)
    labels = utility.read_json_file(labels_filename)
    img_arr = itk.GetArrayViewFromImage(img)

    for label in labels:
        l = labels[label]["label"]
        if "sphere" in label and "shell" not in label:
            img_arr[img_arr == l] = activities[label]
        else:
            img_arr[img_arr == l] = 0

    # The coordinate system is different from IEC analytical volume
    # 35mm should be added in Y
    itk.imwrite(img, source_filename)



#!/usr/bin/env python3

# #################################################################
# This software is distributed under the terms
# of the GNU Lesser General Public Licence (LGPL)
# Hermann Fuchs, Medical University of Vienna / Mayo Clinic Florida
# 10.08.2023
# #################################################################

import opengate as gate
import pathlib
import itk
import numpy as np
import matplotlib.pyplot as plot
from scipy.spatial.transform import Rotation

import opengate.contrib.beamlines.ionbeamline as ionbeamline
import opengate.contrib.tps.ionbeamtherapy as ionbeamtherapy

# from opengate.helpers_rt_plan import *
# from opengate.geometry.helpers_geometry import *
import logging
import argparse
import os
from opengate.tests import utility


# units
km = gate.g4_units.km
m = gate.g4_units.m
mm = gate.g4_units.mm
cm = gate.g4_units.cm
eV = gate.g4_units.eV
MeV = gate.g4_units.MeV
um = gate.g4_units.um
nm = gate.g4_units.nm
km = gate.g4_units.km
gcm3 = gate.g4_units.g / gate.g4_units.cm3
deg = gate.g4_units.deg
mrad = gate.g4_units.mrad


def set_source(sim, n_particles_thread, source_type_str, tp_path):

    ## ---------- DEFINE BEAMLINE MODEL -------------##
    MAbeamline = ionbeamline.BeamlineModel()
    MAbeamline.name = None
    MAbeamline.radiation_types = "ion 6 12"
    # Nozzle entrance to Isocenter distance
    MAbeamline.distance_nozzle_iso = 460.00  # mm
    # SMX to Isocenter distance
    MAbeamline.distance_stearmag_to_isocenter_x = 6700.00
    # SMY to Isocenter distance
    MAbeamline.distance_stearmag_to_isocenter_y = 7420.00
    # polinomial coefficients
    MAbeamline.energy_mean_coeffs = [
        -1.3224984201351649e-08,
        1.7894153922222887e-05,
        -0.008364312581777621,
        13.579959597227932,
        -149.53103267823278,
    ]
    MAbeamline.energy_spread_coeffs = [0.0001]
    MAbeamline.sigma_x_coeffs = [2.0]
    MAbeamline.theta_x_coeffs = [0.0002]
    MAbeamline.epsilon_x_coeffs = [0.0002]
    MAbeamline.sigma_y_coeffs = [2.0]
    MAbeamline.theta_y_coeffs = [0.0002]
    MAbeamline.epsilon_y_coeffs = [0.0002]
    ## --------START PENCIL BEAM SCANNING---------- ##
    # NOTE: HBL means that the beam is coming from -x (90 degree rot around y)
    spots, ntot, energies, G = ionbeamtherapy.spots_info_from_txt(
        tp_path, "ion 6 12", beam_nr=1
    )
    print(dir(spots))
    print(f"{G = }")
    if "tps" in source_type_str.lower():
        # print("sum_weights: ", sum_weights, " sum_beamFraction: ", sum_beamFraction)
        tps = ionbeamtherapy.TreatmentPlanSource("RT_plan", sim)
        tps.set_beamline_model(MAbeamline)
        tps.set_particles_to_simulate(n_particles_thread)
        tps.set_spots(spots)
        tps.rotation = Rotation.from_euler("z", G, degrees=True)
        tps.initialize_tpsource()

        actual_sim_particles = tps.actual_sim_particles

    else:
        set_gps(spots, n_particles_thread, sim, beamline=MAbeamline, gantry_angle=G)


def set_gps(spots, nSim, sim, beamline, gantry_angle):
    for i, spot in enumerate(spots):

        # simulate a fraction of the beam particles for this spot
        nspot = np.round(spot.beamFraction * nSim)

        if nspot == 0:
            continue

        source = sim.add_source("GenericSource", f"gps_spot_{i}")

        # set energy
        source.energy.type = "gauss"
        source.energy.mono = beamline.get_energy(nominal_energy=spot.energy)
        source.energy.sigma_gauss = beamline.get_sigma_energy(
            nominal_energy=spot.energy
        )
        # print(spot.particle_name)
        source.particle = spot.particle_name
        source.position.type = "disc"  # pos = Beam, shape = circle + sigma

        # # set mother
        # if self.mother is not None:
        #     source.attached_to = self.mother

        # # POSITION:
        # source.position.translation = self._get_pbs_position(spot)

        # # ROTATION:
        # source.position.rotation = self._get_pbs_rotation(spot)
        # source.position.rotation = Rotation.from_euler("z", gantry_angle, degrees=True).as_matrix()
        # source.position.radius = 8 * mm
        source.position.sigma_x = 3 * mm
        source.position.sigma_y = 3 * mm
        source.position.translation = [250 * mm, 0 * mm, 0 * mm]
        source.direction.type = "momentum"
        source.direction.momentum = [-1, 0, 0]

        # add weight
        # source.weight = -1
        source.n = nspot

        # set optics parameters
        # source.direction.partPhSp_x = [
        #     beamline.get_sigma_x(spot.energy),
        #     beamline.get_theta_x(spot.energy),
        #     beamline.get_epsilon_x(spot.energy),
        #     beamline.conv_x,
        # ]
        # source.direction.partPhSp_y = [
        #     beamline.get_sigma_y(spot.energy),
        #     beamline.get_theta_y(spot.energy),
        #     beamline.get_epsilon_y(spot.energy),
        #     beamline.conv_y,
        # ]


def simulation(
    data_path,
    output_path,
    number_of_particles=1,
    tp_path="",
    source_type_str="tps",
    calc_LETd=False,
):
    """Run Gate simulation"""
    number_of_threads_for_simulation = os.cpu_count()
    n_particles_thread = number_of_particles / number_of_threads_for_simulation

    # simulation object
    sim = gate.Simulation()
    ui = sim.user_info
    # print(ui)

    ui.running_verbose_level = 0
    # ui.running_verbose_level = 0

    ui.g4_verbose = False
    ui.g4_verbose_level = 0

    # Visualization
    ui.visu = False
    # ui.visu_type = "vrml"
    ui.visu_verbose = False

    ui.random_engine = "MersenneTwister"
    ui.random_seed = "auto"
    # self.random_seed = 123456789
    ui.number_of_threads = number_of_threads_for_simulation

    # Materials
    sim.volume_manager.add_material_database(data_path / "GateMaterials.db")

    # ######################################################################
    # # Defining geometry world plus patient
    # ######################################################################
    # geometry
    # There is a default volume called world (lowercase)
    #  change world size
    world = sim.world
    world.size = [11 * m, 5 * m, 5 * m]
    world.material = "Air"

    # adding nozzle
    # nozzle.create_gantry_nozzle(sim)
    # nozzle.create_FX_nozzle(sim, nozzle_position=100)
    print("Volume tree:")
    print(sim.volume_manager.dump_volume_tree())

    patient = sim.add_volume("Box", "patient")
    print(patient)  # to display the default parameter values
    patient.material = "G4_WATER"
    patient.mother = "world"  # by default
    patient.size = [500 * mm, 400 * mm, 400 * mm]
    # # rotation of the patient
    rotation_matrix = Rotation.from_euler("z", 180, degrees=True).as_matrix()
    patient.rotation = rotation_matrix

    # ######################################################################
    # # Defining actors
    # ######################################################################
    base_name = f"SOBP_{source_type_str}_{number_of_particles:.0f}"
    # Statistics Actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    new_joined_path = os.path.join(output_path, base_name + "Statistics" + ".txt")
    stats.output = new_joined_path
    stats.track_types_flag = True

    # calculate actor size and spacing based on patient size and resolution
    target_voxel_size = [1.0 * mm, patient.size[1] * mm, patient.size[2] * mm]
    # target_voxel_size = [1 * mm, 1 * mm, 1 * mm]

    is_dimensions = [
        int(patient.size[0] / target_voxel_size[0]),
        int(patient.size[1] / target_voxel_size[1]),
        int(patient.size[2] / target_voxel_size[2]),
    ]
    # print("is_dimensions", is_dimensions)
    # print("is_voxel_size", is_voxel_size)

    # ###############################
    # # Dose Actor
    dose = sim.add_actor("DoseActor", "dose")
    # dose.output_filename = output_path / "OneDimDose.mhd"
    dose.attached_to = "patient"
    # dose.spacing = is_voxel_size
    dose.spacing = target_voxel_size
    # print("dose.spacing", dose.spacing)
    dose.size = is_dimensions

    # # number of voxels per dimension
    # dose.size = [517, 460, 228]
    # # size of the voxels
    # dose.spacing = [0.9765625 * mm, 0.9765625 * mm, 2 * mm]
    dose.uncertainty = False
    dose.use_more_ram = False
    dose.hit_type = "random"

    dose.dose = False
    new_joined_path = os.path.join(output_path, base_name + "Dose" + ".mhd")
    dose.output_filename = new_joined_path

    if calc_LETd:
        # # LET Actor
        letactor = sim.add_actor("LETActor", "let")
        # letactor.output = output_path / "OneDimletactor.mhd"
        letactor.mother = "patient"
        # letactor.spacing = is_voxel_size
        letactor.spacing = target_voxel_size
        # print("letactor.spacing", letactor.spacing)
        letactor.size = is_dimensions
        letactor.dose_average = True
        new_joined_path = os.path.join(output_path, base_name + "LET" + ".mhd")
        letactor.output = new_joined_path
        dir(letactor)

    # ######################################################################
    # # Defining pencil beam source
    # ######################################################################

    set_source(sim, n_particles_thread, source_type_str, tp_path)

    # ######################################################################
    # # Defining physics
    # ######################################################################

    # sim.physics_manager.physics_list_name = "FTFP_BERT"
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"

    global_cut = 100 * cm
    sim.physics_manager.global_production_cuts.gamma = global_cut
    sim.physics_manager.global_production_cuts.electron = global_cut
    sim.physics_manager.global_production_cuts.positron = global_cut
    sim.physics_manager.global_production_cuts.proton = global_cut

    # # adding cuts for waterphantom
    # reg = sim.physics_manager.add_region("reg")
    # reg.max_step_size = 0.1 * mm
    # reg.production_cuts.gamma = 0.1 * mm
    # reg.production_cuts.electron = 0.1 * mm
    # reg.production_cuts.positron = 0.1 * mm
    # reg.production_cuts.proton = 0.1 * mm
    # # reg.associate_volume("waterbox")
    # reg.associate_volume("patient")

    # sim.physics_manager.set_user_limits_particles("all")

    # # sim.initialize()
    # print("Phys list cuts to be used: ", sim.physics_manager.dump_production_cuts())
    # # print("info physics lists ", sim.physics_manager.dump_info_physics_lists())
    # print(sim.physics_manager)
    # print("Volume tree:")
    # print(sim.volume_manager.dump_volume_tree())

    # print("Number of sources: ", sim.dump_sources())

    # # start simulation
    # # ~ output = sim.start()
    output = sim.run(start_new_process=True)
    dose_fpath = sim.get_actor("dose").phsp_source.output
    if calc_LETd:
        let_fpath = sim.get_actor("let").phsp_source.output
    else:
        let_fpath = ""
    return output, dose_fpath, let_fpath


def eval_data(dose_ref_fpath, dose_fpath, scale_img2=1.0):
    utility.assert_images(
        dose_ref_fpath,
        dose_fpath,
        axis="x",
        scaleImageValuesFactor=scale_img2,
    )


def main():
    # logging.basicConfig(level=logging.DEBUG)
    logging.basicConfig(level=logging.INFO)
    print("Running Gate Simulation")
    current_path = pathlib.Path(__file__).parent.resolve()
    data_path = current_path / "../data"
    output_path = current_path / "../temp"
    # number_of_particles = 100000
    number_of_particles_1 = 1e4
    source_type_str = "gps"
    calc_LETd = False
    tp_path = data_path / "tpsource/SOBP_HFU_23_02_2024.txt"
    simLink, dose_1e5_fpath, let_fpath_1e5 = simulation(
        data_path,
        output_path,
        number_of_particles=number_of_particles_1,
        tp_path=tp_path,
        source_type_str=source_type_str,
        calc_LETd=calc_LETd,
    )
    number_of_particles_2 = 1.0e5

    simLink, dose_fpath, let_fpath = simulation(
        data_path,
        output_path,
        number_of_particles=number_of_particles_2,
        tp_path=tp_path,
        source_type_str=source_type_str,
        calc_LETd=calc_LETd,
    )
    scale_img2 = number_of_particles_1 / number_of_particles_2
    eval_data(dose_1e5_fpath, dose_fpath, scale_img2=scale_img2)
    if calc_LETd:
        eval_data(let_fpath_1e5, let_fpath)
    print("Simulation finished")


if __name__ == "__main__":
    main()



import wget
import os
import tarfile
import platform
import sys
from .g4DataSetup import *


# Set Qt5 data paths:
def set_qt5_path():
    s = platform.system()
    if s == "Linux" or s == "Windows":
        g4libFolder = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "..", "opengate_core.libs"
        )
    elif s == "Darwin":
        g4libFolder = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), ".dylibs"
        )

    os.environ["QTHOME"] = ""
    os.environ["QTLIBPATH"] = g4libFolder
    os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "plugins"
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test037_pet_hits_singles_helpers as t37
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test037_pet", "test037")

    # test version
    v = "2_2"

    # create the simulation
    sim = gate.Simulation()
    crystal = t37.create_pet_simulation(sim, paths, create_mat=True)
    module = sim.volume_manager.volumes["pet_module"]
    die = sim.volume_manager.volumes["pet_die"]
    stack = sim.volume_manager.volumes["pet_stack"]

    # digitizer hits
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.attached_to = crystal.name
    hc.authorize_repeated_volumes = True
    hc.output_filename = f"test037_test{v}.root"
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # Readout (not need for adder)
    sc = sim.add_actor("DigitizerReadoutActor", "Singles2_1")
    sc.authorize_repeated_volumes = True
    sc.output_filename = f"test037_test{v}.root"
    sc.input_digi_collection = "Hits"
    sc.group_volume = stack.name  # should be depth=1 in Gate
    sc.discretize_volume = crystal.name
    sc.policy = "EnergyWeightedCentroidPosition"

    # Readout: another one, with different option (in the same output file)
    sc = sim.add_actor("DigitizerReadoutActor", "Singles2_2")
    sc.output_filename = f"test037_test{v}.root"
    sc.input_digi_collection = "Hits"
    sc.group_volume = crystal.name  # should be depth=4 in Gate
    sc.discretize_volume = crystal.name
    sc.policy = "EnergyWeightedCentroidPosition"

    # timing
    sec = gate.g4_units.second
    sim.run_timing_intervals = [[0, 0.00005 * sec]]

    # start simulation
    sim.run()

    # print results
    stats = sim.get_actor("Stats")
    print(stats)

    # ----------------------------------------------------------------------------------------------------------

    # check stats
    print()
    gate.exception.warning(f"Check stats")
    # p = paths.gate / "output_test1"
    p = paths.gate / "output"
    stats_ref = utility.read_stat_file(p / f"stats{v}.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.03)

    # check root hits
    hc = sim.get_actor("Hits")
    f = p / f"output{v}.root"
    is_ok = t37.check_root_hits(paths, v, f, hc.get_output_path()) and is_ok

    # check root singles
    sc = sim.get_actor("Singles2_1")
    f = p / f"output2_1.root"
    is_ok = (
        t37.check_root_singles(paths, "2_1", f, sc.get_output_path(), sc.name) and is_ok
    )

    # check root singles
    sc = sim.get_actor("Singles2_2")
    f = p / f"output2_2.root"
    is_ok = (
        t37.check_root_singles(paths, "2_2", f, sc.get_output_path(), sc.name) and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import pathlib
import os
import subprocess


# This test is to be run after test069_sim_as_dict.py
# It checks whether the simulation is recreated correctly from the JSON file,
# but does not actually run any simulation.
def main(a, b, c, dependency="test069_sim_as_dict.py"):
    pathFile = pathlib.Path(__file__).parent.resolve()
    paths = utility.get_default_test_paths(__file__)

    # the test069_sim_as_dict.py is needed first
    if dependency and not os.path.isfile(paths.output / "simu_test069.json"):
        print(f"Running test069_sim_as_dict.py")
        subprocess.call(["python", paths.current / dependency])

    # create the simulation
    sim = gate.create_sim_from_json(paths.output / "test069" / "simu_test069.json")

    # Assert that objects have been read back correctly
    assert "rod" in sim.volume_manager.volumes
    m = gate.g4_units.m
    assert sim.volume_manager.world_volume.size == [1.5 * m, 1.5 * m, 1.5 * m]
    assert "rod_region" in sim.physics_manager.regions
    assert pathlib.Path(paths.output / "test069" / "patient-4mm.mhd").exists()
    assert (
        sim.volume_manager.get_volume("waterbox_with_hole").creator_volumes[0].name
        == "Waterbox"
    )

    # If we make it until here without exception, the test is passed
    utility.test_ok(True)


if __name__ == "__main__":
    main(1, 2, 3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test023")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 6549
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    mm = gate.g4_units.mm

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 50 * MeV
    source.particle = "proton"
    source.position.type = "sphere"
    source.position.radius = 1 * cm
    source.direction.type = "iso"
    source.activity = 30000 * Bq

    # filter : keep e- only
    fp = sim.add_filter("ParticleFilter", "fp")
    fp.particle = "e-"

    # add dose actor
    dose1 = sim.add_actor("DoseActor", "dose1")
    dose1.output_filename = "test023.mhd"
    dose1.attached_to = waterbox
    dose1.size = [100, 100, 100]
    dose1.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose1.filters.append(fp)

    # add dose actor, without e- (to check)
    fe = sim.add_filter("ParticleFilter", "fe")
    fe.particle = "e-"
    fe.policy = "reject"
    print(dir(fe))

    dose2 = sim.add_actor("DoseActor", "dose2")
    dose2.output_filename = "test023-noe.mhd"
    dose2.attached_to = waterbox
    dose2.size = [100, 100, 100]
    dose2.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose2.filters.append(fe)

    """fe = sim.add_filter("ParticleFilter", "f")
    fe.particle = "gamma"
    fe.policy = "reject"
    dose2.filters.append(fe)"""

    # add stat actor (only gamma)
    fg = sim.add_filter("ParticleFilter", "fg")
    fg.particle = "gamma"

    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True
    stat.filters.append(fg)

    print("Filters: ", sim.filter_manager)
    print(sim.filter_manager.dump())

    # change physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.physics_manager.global_production_cuts.all = 0.1 * mm

    # start simulation
    sim.run()

    # print results at the end
    print(stat)
    print(dose1)
    print(dose2)

    # tests
    stats_ref = utility.read_stat_file(paths.output_ref / "test023_stats.txt")
    is_ok = utility.assert_stats(stat, stats_ref, 0.8)

    print()
    is_ok = is_ok and utility.assert_images(
        paths.output_ref / "test023-edep.mhd",
        dose1.edep.get_output_path(),
        stat,
        tolerance=50,
        sum_tolerance=4,
    )

    print()
    is_ok = is_ok and utility.assert_images(
        paths.output_ref / "test023-noe-edep.mhd",
        dose2.edep.get_output_path(),
        stat,
        tolerance=40,
        sum_tolerance=2,
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
import SimpleITK as sitk
import subprocess
import os, sys


def main(dependency="test085_free_flight_spect_2b_fd_mt.py"):

    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test085_spect"
    )

    # not on windows
    if os.name == "nt":
        sys.exit(0)

    # The test needs the output of the other tests
    if not os.path.isfile(paths.output / "projection_1_fd_counts.mhd"):
        subprocess.call(["python", paths.current / dependency])
    if not os.path.isfile(paths.output / "projection_1_ff_sc_counts.mhd"):
        subprocess.call(
            ["python", paths.current / "test085_free_flight_spect_3_scatter_mt.py"]
        )

    # ratios of the number of particles
    ref_n = 5e6
    prim_n = 6e3
    sca_n = 1e4
    scaling_prim = ref_n / prim_n
    scaling_sc = ref_n / sca_n

    # merge primary and scatter
    prim_filename = paths.output / f"projection_1_fd_counts.mhd"
    sca_filename = paths.output / f"projection_1_ff_sc_counts.mhd"

    # read images
    prim_img = sitk.ReadImage(prim_filename)
    sca_img = sitk.ReadImage(sca_filename)

    print("Number of events, ref = ", ref_n)
    print("Number of events, fd = ", prim_n, scaling_prim)
    print("Number of events, sc = ", sca_n, scaling_sc)

    # sum both images
    arr = (
        sitk.GetArrayFromImage(prim_img) * scaling_prim
        + sitk.GetArrayFromImage(sca_img) * scaling_sc
    )
    img = sitk.GetImageFromArray(arr)
    img.CopyInformation(prim_img)
    sitk.WriteImage(img, paths.output / f"projection_1_total.mhd")

    # compare to noFF
    is_ok = True

    print()
    print("fd counts must be lower than ref counts")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "projection_1_ref_counts.mhd",
            paths.output / "projection_1_fd_counts.mhd",
            None,
            tolerance=100,
            ignore_value_data1=0,
            sum_tolerance=45,
            axis="x",
            scaleImageValuesFactor=scaling_prim,
            slice_id=1,
            fig_name=paths.output / f"projection_1_ff_test_1.png",
        )
        and is_ok
    )

    print()
    is_ok = (
        utility.assert_images(
            paths.output_ref / "projection_1_ref_counts.mhd",
            paths.output / "projection_1_ff_sc_counts.mhd",
            None,
            tolerance=180,
            ignore_value_data1=0,
            sum_tolerance=48,
            axis="x",
            scaleImageValuesFactor=scaling_sc,
            slice_id=0,
            fig_name=paths.output / f"projection_1_ff_sc_test_1.png",
        )
        and is_ok
    )

    print()
    print(f"Compare the head n1, slice scatter")
    is_ok = (
        utility.assert_images(
            paths.output_ref / f"projection_1_ref_counts.mhd",
            paths.output / f"projection_1_total.mhd",
            None,
            tolerance=180,
            ignore_value_data1=0,
            sum_tolerance=48,
            axis="x",
            sad_profile_tolerance=35,
            slice_id=0,
            fig_name=paths.output / f"projection_1_total_test_scatter.png",
        )
        and is_ok
    )

    print()
    print(f"Compare the head n1, slice primary")
    is_ok = (
        utility.assert_images(
            paths.output_ref / f"projection_1_ref_counts.mhd",
            paths.output / f"projection_1_total.mhd",
            None,
            tolerance=71,
            ignore_value_data1=0,
            sum_tolerance=20,
            axis="x",
            sad_profile_tolerance=24,
            slice_id=1,
            fig_name=paths.output / f"projection_1_total_test_prim.png",
        )
        and is_ok
    )

    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test060_phsp_source_helpers as t
import opengate as gate
import uproot
import numpy as np
import gatetools.phsp as phsp
from opengate.tests import utility


paths = utility.get_default_test_paths(
    __file__, "test060_PhsSource_ParticleName_direct", output_folder="test060"
)

# units
m = gate.g4_units.m
mm = gate.g4_units.mm
cm = gate.g4_units.cm
nm = gate.g4_units.nm
Bq = gate.g4_units.Bq
MeV = gate.g4_units.MeV
deg: float = gate.g4_units.deg


def create_root_file(file_name):
    # create numpy arrays
    kinetic_energy = np.array([100, 2.0, 3.0, 80, 2.0, 3.0, 2.0, 110.0, 3.0, 111.0, 2])
    pdg_code = np.array([2212, 11, 11, 2212, 11, 11, 11, 2212, 11, 2212, 11])
    pre_direction_local_x = np.zeros(11)
    pre_direction_local_y = np.zeros(11)
    pre_direction_local_z = np.ones(11)
    pre_position_local_x = np.zeros(11)
    pre_position_local_y = np.zeros(11)
    pre_position_local_z = np.zeros(11)
    pre_direction_x = np.zeros(11)
    pre_direction_y = np.zeros(11)
    pre_direction_z = np.ones(11)
    pre_position_x = np.zeros(11)
    pre_position_y = np.zeros(11)
    pre_position_z = np.zeros(11)
    weight = np.ones(11)

    # generate root file
    tfile = uproot.recreate(file_name)
    tfile["PhaseSpace1"] = {
        "KineticEnergy": kinetic_energy,
        "PDGCode": pdg_code,
        "PreDirectionLocal_X": pre_direction_local_x,
        "PreDirectionLocal_Y": pre_direction_local_y,
        "PreDirectionLocal_Z": pre_direction_local_z,
        "PrePositionLocal_X": pre_position_local_x,
        "PrePositionLocal_Y": pre_position_local_y,
        "PrePositionLocal_Z": pre_position_local_z,
        "PreDirection_X": pre_direction_x,
        "PreDirection_Y": pre_direction_y,
        "PreDirection_Z": pre_direction_z,
        "PrePosition_X": pre_position_x,
        "PrePosition_Y": pre_position_y,
        "PrePosition_Z": pre_position_z,
        "Weight": weight,
    }
    tfile.close()


def main():
    print("create reference PhS file")
    create_root_file(paths.output / "test_phs.root")

    print("testing until primary")
    t.test_source_until_primary(
        source_file_name=paths.output / "test_phs.root",
        phs_file_name_out=paths.output / "test_source_untilPrimary.root",
    )

    # load data from root file
    data_ref, keys_ref, m_ref = phsp.load(
        paths.output / "test_source_untilPrimary.root"
    )
    # the root file should contain 9 particles
    # print(m_ref)

    # there should be 3 protons in the root file
    index = keys_ref.index("PDGCode")
    particle_list = data_ref[:, index]
    # print(particle_list)
    count = np.count_nonzero(particle_list == 2212)
    is_ok = False
    if m_ref == 9 and count == 3:
        is_ok = True
        print("test ok")

    # this is the end, my friend
    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import uproot
import numpy as np
import matplotlib.pyplot as plt
import os
import subprocess
import gatetools.phsp as phsp

import opengate as gate
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.tests import utility

global all_cond


def main(dependency="test040_gan_phsp_pet_aref.py"):
    global all_cond
    paths = utility.get_default_test_paths(__file__, "", "test040")

    # The test needs the output of test040_gan_phsp_pet_aref.py
    # If the output of test040_gan_phsp_pet_aref.py does not exist (eg: random test), create it
    if not os.path.isfile(paths.output / "test040_gan_phsp.root"):
        print("---------- Begin of test040_gan_phsp_pet_aref.py ----------")
        subprocess.call(["python", paths.current / dependency])
        print("----------- End of test040_gan_phsp_pet_aref.py -----------")

    # create the simulation
    sim = gate.Simulation()

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    cm3 = gate.g4_units.cm3
    keV = gate.g4_units.keV
    MeV = gate.g4_units.MeV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    BqmL = Bq / cm3
    sec = gate.g4_units.second
    deg = gate.g4_units.deg
    kBq = 1000 * Bq
    MBq = 1000 * kBq

    # main parameters
    sim.output_dir = paths.output
    sim.check_volumes_overlap = True
    sim.number_of_threads = 1
    sim.random_seed = 123456
    # sim.running_verbose_level = gate.EVENT
    # sim.g4_verbose = True
    sim.output_dir = paths.output
    ac = 5e3 * BqmL / sim.number_of_threads
    sim.visu = False
    colli_flag = not sim.visu
    if sim.visu:
        ac = 1 * BqmL
        sim.number_of_threads = 1

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # test phase space to check with reference
    phsp_sphere_surface = sim.add_volume("Sphere", "phase_space_sphere")
    phsp_sphere_surface.rmin = 215 * mm
    phsp_sphere_surface.rmax = 216 * mm
    phsp_sphere_surface.color = [1, 1, 1, 1]
    phsp_sphere_surface.material = "G4_AIR"

    # physic list
    sim.physics_manager.set_production_cut("world", "all", 1 * mm)

    # activity parameters
    spheres_diam = [10, 13, 17, 22, 28, 37]
    spheres_activity_concentration = [ac * 6, ac * 5, ac * 4, ac * 3, ac * 2, ac]

    # initialisation for conditional
    spheres_radius = [x / 2.0 for x in spheres_diam]
    spheres_centers, spheres_volumes = gate_iec.get_default_sphere_centers_and_volumes()
    spheres_activity_ratio = []
    spheres_activity = []
    for diam, ac, volume, center in zip(
        spheres_diam, spheres_activity_concentration, spheres_volumes, spheres_centers
    ):
        activity = ac * volume
        print(
            f"Sphere {diam}: {str(center):<30} {volume / cm3:7.3f} cm3 "
            f"{activity / Bq:7.0f} Bq  {ac / BqmL:7.1f} BqmL"
        )
        spheres_activity.append(activity)

    total_activity = sum(spheres_activity)
    print(f"Total activity {total_activity / Bq:.0f} Bq")
    for activity in spheres_activity:
        spheres_activity_ratio.append(activity / total_activity)
    print("Activity ratio ", spheres_activity_ratio, sum(spheres_activity_ratio))

    # will store all conditional info (position, direction)
    all_cond = None

    # unique (reproducible) random generator
    rs = gate.utility.get_rnd_seed(123456)

    # FIXME -> should not be in the main ; warning used global variable (spheres_activity_ratio)
    def gen_cond(n):
        n_samples = gate_iec.get_n_samples_from_ratio(n, spheres_activity_ratio)
        # (it is required to shuffle when using several activity spheres to avoid time artifact)
        cond = gate_iec.generate_pos_spheres(
            spheres_centers, spheres_radius, n_samples, shuffle=True, rs=rs
        )
        # we keep all conditions for the test (not needed in normal simulation)
        global all_cond
        if all_cond is None:
            all_cond = cond
        else:
            all_cond = np.column_stack((all_cond, cond))

        return cond

    """
        train the GAN

    1 - generate the training dataset with test040
        5 MB pour 15 BqmL, 1 thread, 4 sec linux
        for 1500 *2 on 4 threads = 718 sec and 3.7 GB

    2 - convert root to npy pairs
        gaga_pet_to_pairs test040_train_big.root -o test040_train_big.npy
        gaga_pairs_to_tlor test040_train_big.npy -o test040_train_big_tlor.npy
        final size is 1.3 GB

    3 - train
        gaga_train test040_train_small_tlor.npy test9221_v4.json -pi epoch 10 -f . -p penalty_weight 10 -ps penalty GP_0GP
        gaga_train test040_train_big_tlor.npy test9221_v4.json -pi epoch 10 -f . -p penalty_weight 10 -ps penalty GP_0GP
        (was "pth120_test9221_GP_0GP_10.0_100000.pth")
        new is test9221_GP_0GP_10.0_100000.pth

    4 - on JZ (see scripts)
    """

    # GAN source
    gsource = sim.add_source("GANPairsSource", "gaga")
    gsource.particle = "gamma"
    # no phantom, we consider attached to the world at origin
    gsource.activity = total_activity
    # gsource.pth_filename = paths.data / "pth120_test9221_GP_0GP_10.0_100000.pth"
    gsource.pth_filename = paths.data / "test9221_GP_0GP_10.0_100000.pth"
    gsource.position_keys = ["X1", "Y1", "Z1", "X2", "Y2", "Z2"]
    gsource.direction_keys = ["dX1", "dY1", "dZ1", "dX2", "dY2", "dZ2"]
    gsource.energy_key = ["E1", "E2"]
    gsource.time_key = ["t1", "t2"]
    # time is added to the simulation time
    gsource.relative_timing = True
    gsource.weight_key = None
    # particle are move backward with 10 cm
    gsource.backward_distance = 10 * cm
    gsource.backward_force = True
    # if the kinetic E is below this threshold, we set it to 0
    gsource.energy_min_threshold = 0.1 * keV
    gsource.energy_max_threshold = 1 * MeV
    gsource.skip_policy = "ZeroEnergy"
    gsource.batch_size = 1e5
    gsource.verbose_generator = True
    # set the generator and the condition generator
    gsource.generator = gate.sources.gansources.GANSourceConditionalPairsGenerator(
        gsource, 210 * mm, gen_cond
    )
    gsource.gpu_mode = (
        utility.get_gpu_mode_for_tests()
    )  # should be "auto" but "cpu" for macOS github actions to avoid mps errors

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.output_filename = "test040_gan_stats.txt"

    # phsp actor
    phsp_actor = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp_actor.attached_to = phsp_sphere_surface.name
    phsp_actor.attributes = [
        "KineticEnergy",
        "PrePosition",
        "PreDirection",
        "GlobalTime",
        "EventPosition",
        "EventDirection",
        "TimeFromBeginOfEvent",
        "EventKineticEnergy",
    ]
    phsp_actor.output_filename = "test040_gan_phsp.root"
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    phsp_actor.filters.append(f)
    f = sim.add_filter("KineticEnergyFilter", "f2")
    f.energy_min = 100 * keV
    phsp_actor.filters.append(f)

    # ----------------------------------------------------------------------------------------------
    # FIXME: cannot be spawn in another process !
    # sim.running_verbose_level = gate.EVENT
    sim.run(start_new_process=False)

    # ----------------------------------------------------------------------------------------------
    # print stats
    print()
    gate.exception.warning(f"Check stats")
    if sim.number_of_threads == 1:
        s = sim.source_manager.get_source("gaga")
    else:
        s = sim.source_manager.get_source_mt("gaga", 0)
    print(f"Source, nb of skipped particles : {s.GetTotalSkippedEvents()}")
    b = gsource.total_skipped_events
    print(f"Source, nb of skipped particles (check) : {b}")

    print(f"Source, nb of zerosE particles : {s.total_zero_events}")
    b = gsource.total_zero_events
    print(f"Source, nb of zerosE particles (check) : {b}")

    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "test040_ref_stats.txt")
    r = (stats_ref.counts.steps - stats.counts.steps) / stats_ref.counts.steps
    print(f"!!! Steps cannot be compared => was {stats.counts.steps}, {r:.2f}%")
    stats.counts.steps = stats_ref.counts.steps
    r = (stats_ref.counts.tracks - stats.counts.tracks) / stats_ref.counts.tracks
    print(f"!!! Tracks cannot be compared => was {stats.counts.tracks}, {r:.2f}%")
    stats.counts.tracks = stats_ref.counts.tracks
    is_ok = utility.assert_stats(stats, stats_ref, 0.10)

    # save conditional for checking with reference cond
    keys = ["EventPosition_X", "EventPosition_Y", "EventPosition_Z"]
    phsp.save_npy(paths.output / "test040_gan_phsp_cond.npy", all_cond, keys)

    # ----------------------------------------------------------------------------------------------
    # compare conditional
    # less particle in the ref because conditional data are stored
    # when exit (not absorbed)
    print()
    gate.exception.warning(f"Check conditions (position, direction)")
    # look root generated by previous 40_aref test
    root_ref = paths.output / "test040_ref_phsp.root"
    hits1 = uproot.open(root_ref)
    branch = hits1.keys()[0]
    print("Branch name:", branch)
    hits1 = hits1[branch]
    names = [k for k in hits1.keys()]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")

    # in the ref phsp, EventID is not unique (pairs of gamma), we set them unique
    # to be compared to the generated cond
    event_id = hits1["EventID"]
    print("Nb of event (non unique)", event_id.shape)
    u, indices = np.unique(event_id, return_index=True)
    for k in hits1:
        hits1[k] = hits1[k][indices]
    event_id = hits1["EventID"]
    print("Nb of event (unique)", event_id.shape)

    root_gan = paths.output / "test040_gan_phsp_cond.npy"
    hits2, hits2_keys, hits2_n = phsp.load(root_gan)
    tols = [10.0] * len(keys)
    tols[keys.index("EventPosition_X")] = 0.21
    # FIXME warning : there is a shift in Y because the pth was done
    # before IEC phantom was corrected. Need to redo the GAN.
    # In the meantime, increase the tol
    tols[keys.index("EventPosition_Y")] = 0.2
    tols[keys.index("EventPosition_Z")] = 0.15
    scalings = [1] * len(keys)
    is_ok = (
        utility.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2_keys),
            keys,
            keys,
            tols,
            scalings,
            scalings,
            True,
        )
        and is_ok
    )
    # figure
    img_filename = paths.output / "test040_cond.png"
    plt.suptitle(
        f"Values: ref {os.path.basename(root_ref)} {os.path.basename(root_gan)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(img_filename)
    print(f"Figure in {img_filename}")

    # ----------------------------------------------------------------------------------------------
    # compare output phsp
    # absorbed events are when E == 0
    print()
    gate.exception.warning(
        f"Check output phsp (the pth is not perfect, but it should be sufficient for tests"
    )
    print("Warning: in the ref phsp, we remove when E==0 (absorbed events)")

    ref_file = paths.output / "test040_ref_phsp.root"
    hits1 = uproot.open(root_ref)
    branch = hits1.keys()[0]
    print("Branch name:", branch)
    hits1 = hits1[branch]
    names = [k for k in hits1.keys()]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")

    # in the ref, remove when E == 0
    ke = hits1["KineticEnergy"]
    print("Nb of event (all E)", ke.shape)
    mask = ke > 0
    for k in hits1:
        hits1[k] = hits1[k][mask]
    ke = hits1["KineticEnergy"]
    print("Nb of event (non E==0)", ke.shape)

    hc_file = phsp_actor.get_output_path()
    hits2, hits2_keys, hits2_n = phsp.load(hc_file)

    checked_keys = [
        "GlobalTime",
        "KineticEnergy",
        "PrePosition_X",
        "PrePosition_Y",
        "PrePosition_Z",
        "PreDirection_X",
        "PreDirection_Y",
        "PreDirection_Z",
    ]
    scalings = [1.0] * len(checked_keys)
    scalings[checked_keys.index("GlobalTime")] = 1e-9  # time in ns
    tols = [10.0] * len(checked_keys)
    tols[checked_keys.index("GlobalTime")] = 0.003
    tols[checked_keys.index("KineticEnergy")] = 0.01
    tols[checked_keys.index("PrePosition_X")] = 2
    tols[checked_keys.index("PrePosition_Y")] = 1
    tols[checked_keys.index("PrePosition_Z")] = 1
    tols[checked_keys.index("PreDirection_X")] = 0.02
    tols[checked_keys.index("PreDirection_Y")] = 0.02
    tols[checked_keys.index("PreDirection_Z")] = 0.02
    is_ok = (
        utility.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2_keys),
            checked_keys,
            checked_keys,
            tols,
            scalings,
            scalings,
            True,
        )
        and is_ok
    )
    # figure
    img_filename = paths.output / "test040_phsp.png"
    plt.suptitle(
        f"Values: ref {os.path.basename(root_ref)} {os.path.basename(root_gan)} "
        f"-> {hits1_n} vs {hits2_n}"
    )
    plt.savefig(img_filename)
    print(f"Figure in {img_filename}")

    # is_ok = utility.compare_root3(ref_file, hc_file, "phsp", "phsp",
    #                          checked_keys, checked_keys, tols, scalings, scalings,
    #                          paths.output / 'test040_phsp.png') and is_ok

    # ----------------------------------------------------------------------------------------------

    # this is the end, my friend
    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test028_ge_nm670_spect_2_helpers as test028
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "", output_folder="test028_proj_blur"
    )

    # create the simulation
    sim = gate.Simulation()

    # main description
    spect = test028.create_spect_simu(sim, paths, 1)
    proj = test028.test_add_proj(sim)

    # change the digitizer to add blurring between the adder and the energy window
    mm = gate.g4_units.mm
    hc = sim.actor_manager.get_actor("Hits")
    sc = sim.actor_manager.get_actor("Singles")
    cc = sim.actor_manager.get_actor("EnergyWindows")

    bc = sim.add_actor("DigitizerSpatialBlurringActor", "Singles_blur")
    bc.output_filename = sc.output_filename
    bc.attached_to = "spect_crystal"  # important !
    bc.input_digi_collection = sc.name
    bc.blur_attribute = "PostPosition"
    bc.use_truncated_Gaussian = False
    # r = 20 * mm
    bc.blur_fwhm = 20 * mm
    bc.keep_in_solid_limits = True

    # input of EnergyWindows is the blur one
    cc.input_digi_collection = "Singles_blur"

    # we modify the priority to have the correct actors order
    hc.priority = 90
    sc.priority = 91
    bc.priority = 92
    cc.priority = 93

    # rotate spect
    cm = gate.g4_units.cm
    test028.nm670.rotate_gantry(spect, 10 * cm, -15)

    sec = gate.g4_units.second
    sim.run_timing_intervals = [[1 * sec, 2 * sec]]

    print(sim)
    sim.run()

    # print stats
    stats = sim.get_actor("Stats")
    print(stats)

    # check singles
    print()
    gate.exception.warning(
        "Compare singles blur vs unblur (it is normal that it dont match)"
    )
    hc_file = sim.get_actor("Singles_blur").get_output_path()
    ref_file = paths.output_ref / hc_file.name
    print(hc_file)
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 3, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 2, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 3, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.1,
            "scaling": 1,
        },
    ]
    is_ok = utility.compare_root2(
        ref_file,
        hc_file,
        "Singles",
        "Singles_blur",
        checked_keys,
        paths.output / f"test028_singles_blur_unblur.png",
    )

    # check singles
    print()
    version = "3_blur"
    gate.exception.warning("Compare singles")
    hc_file = sim.get_actor("Singles_blur").get_output_path()
    ref_file = paths.output_ref / hc_file.name
    print(hc_file)
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 0.9, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 0.3, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.4, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.0002,
            "scaling": 1,
        },
    ]
    is_ok = (
        utility.compare_root2(
            ref_file,
            hc_file,
            "Singles_blur",
            "Singles_blur",
            checked_keys,
            paths.output / f"test028_singles.png",
        )
        and is_ok
    )

    # check projection
    proj_out = sim.get_actor("Projection")
    is_ok = test028.test_spect_proj(sim, paths, proj_out) and is_ok

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test073_helpers import *
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, gate_folder="gate_test073_intevo", output_folder="test073"
    )

    # create the simulation
    sim = gate.Simulation()
    create_sim_tests(sim, threads=4, digitizer=1)
    sim.output_dir = paths.output

    # timing
    sim.random_seed = 123654789
    sec = gate.g4_units.second
    stop = 3 * sec
    if sim.visu:
        stop = 0.001 * sec
    sim.run_timing_intervals = [[0, stop]]

    # output filenames
    stats = sim.get_actor("stats")
    stats.output_filename = "stats1.txt"

    crystal = sim.volume_manager.get_volume(f"spect_crystal")
    proj = sim.actor_manager.find_actor_by_type("DigitizerProjectionActor")
    proj.output_filename = "projections_test1.mhd"

    hits = sim.actor_manager.find_actor_by_type("DigitizerHitsCollectionActor")
    hits.output_filename = "output_test1.root"
    hits.write_to_disk = True

    singles = sim.actor_manager.find_actor_by_type("DigitizerAdderActor")
    singles.output_filename = "output_test1.root"
    singles.write_to_disk = True

    # start simulation
    sim.run()

    # print stats
    print(stats)

    # ------------------------------------------------------------------------------------
    # FIXME: WARNING !!!!
    # This is ** not ** a validation of the SPECT model.
    # This is a FAKE digitizer.
    # This test only checks that the output is consistant with some reference data obtained
    # with an old Gate9 simulation.
    # This test will change in the future, once the validation is completed.
    # FIXME: WARNING !!!!
    # ------------------------------------------------------------------------------------

    # compare stats
    is_ok = compare_stats(sim, paths.gate_output / "stats1.txt")

    # compare root
    fr = paths.gate_output / "output1.root"
    is_ok = compare_root_hits(crystal, sim, fr, paths.output) and is_ok

    # Compare root files
    sn = f"{crystal.name}_singles"
    is_ok = compare_root_singles(crystal, sim, fr, paths.output, sn) and is_ok

    # compare images with Gate
    fr = paths.gate_output / "projection1.mhd"
    is_ok = compare_proj_images(crystal, sim, stats, fr, paths.output) and is_ok
    utility.test_ok(is_ok)



import sys
import inspect

from . import image, listmode, sequences, datafetchers, utility

# import inspect
#
# available_processing_units = {}
# g = globals().copy()
# for name, obj in g.items():
#     if inspect.isclass(obj):
#         available_processing_units[name] = obj
# print(available_processing_units)

available_processing_units = {
    "ProjectionListMode": image.ProjectionListMode,
    "ProcessingSequence": sequences.ProcessingSequence,
    "GaussianBlurringSingleAttribute": listmode.GaussianBlurringSingleAttribute,
    "OffsetSingleAttribute": listmode.OffsetSingleAttribute,
    "DataFetcherHdf5": datafetchers.DataFetcherHdf5,
}

clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
import opengate.contrib.spect.ge_discovery_nm670 as gate_spect
import opengate as gate
import test043_garf_helpers as test43
from opengate.tests import utility

if __name__ == "__main__":
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.number_of_threads = 3
    sim.visu = False
    sim.random_seed = 321654987
    sim.output_dir = test43.paths.output

    # units
    nm = gate.g4_units.nm
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    Bq = gate.g4_units.Bq
    keV = gate.g4_units.keV

    # activity
    activity = 1e6 * Bq / sim.number_of_threads

    # add a material database
    sim.volume_manager.add_material_database(test43.paths.data / "GateMaterials.db")

    # init world
    test43.sim_set_world(sim)

    # fake spect head
    head = gate_spect.add_fake_spect_head(sim, "spect")
    head.translation = [0, 0, -15 * cm]

    # detector input plane (+ 1nm to avoid overlap)
    pos, crystal_dist, psd = gate_spect.get_plane_position_and_distance_to_crystal(
        "lehr"
    )
    pos += 1 * nm
    print(f"plane position     {pos / mm} mm")
    print(f"crystal distance   {crystal_dist / mm} mm")
    detPlane = test43.sim_add_detector_plane(sim, head.name, pos)

    # physics
    test43.sim_phys(sim)

    # sources
    test43.sim_source_test(sim, activity)

    # arf actor
    arf = sim.add_actor("ARFActor", "arf")
    arf.attached_to = detPlane.name
    arf.output_filename = test43.paths.output / "test043_projection_garf_mt.mhd"
    arf.batch_size = 2e5
    arf.image_size = [128, 128]
    arf.image_spacing = [4.41806 * mm, 4.41806 * mm]
    arf.verbose_batch = True
    arf.distance_to_crystal = crystal_dist  # 74.625 * mm
    arf.distance_to_crystal = 74.625 * mm
    # arf.pth_filename = test43.paths.gate_data / "pth" / "arf_Tc99m_v3.pth"
    arf.pth_filename = test43.paths.gate_data / "pth" / "arf_Tc99m_v034.pth"
    arf.enable_hit_slice = True
    arf.flip_plane = True  # because the training was backside
    arf.gpu_mode = (
        utility.get_gpu_mode_for_tests()
    )  # should be "auto" but "cpu" for macOS github actions to avoid mps errors

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # print info
    print("")
    arf = sim.get_actor("arf")
    img = itk.imread(str(arf.get_output_path("counts")))
    # set the first channel to the same channel (spectrum) than the analog
    img[0, :] = img[1, :] + img[2, :]
    print(f"Number of batch: {arf.batch_nb}")
    print(f"Number of detected particles: {arf.detected_particles}")
    filename1 = str(arf.get_output_path("counts")).replace(".mhd", "_0.mhd")
    itk.imwrite(img, filename1)

    # high stat
    filename2 = str(arf.get_output_path("counts")).replace(".mhd", "_hs.mhd")
    scale = 4e8 * Bq / activity / sim.number_of_threads
    print(f"Scaling ref = 4e8, activity = {activity}, scale = {scale}")
    img2 = gate.image.scale_itk_image(img, scale)
    itk.imwrite(img2, filename2)

    # ----------------------------------------------------------------------------------------------------------------
    # tests
    print()
    gate.exception.warning("Tests stats file")
    stats_ref = utility.read_stat_file(test43.paths.gate_output / "stats_analog.txt")
    # dont compare steps of course
    stats_ref.counts.steps = stats.counts.steps
    stats_ref.counts.runs = 3
    is_ok = utility.assert_stats(stats, stats_ref, 0.01)

    print()
    gate.exception.warning("Compare image to analog")
    is_ok = (
        utility.assert_images(
            test43.paths.output_ref / "test043_projection_analog.mhd",
            filename1,
            stats,
            tolerance=100,
            ignore_value_data2=0,
            axis="x",
            sum_tolerance=20,
        )
        and is_ok
    )

    print()
    gate.exception.warning("Compare image to analog high statistics")
    is_ok = (
        utility.assert_images(
            test43.paths.output_ref / "test043_projection_analog_high_stat.mhd",
            filename2,
            stats,
            tolerance=52,
            ignore_value_data2=0,
            axis="x",
        )
        and is_ok
    )

    print()
    gate.exception.warning("profile compare : ")
    p = test43.paths.output_ref
    print(
        f'garf_compare_image_profile {p / "test043_projection_analog.mhd"} {filename1} -w 3'
    )
    print(
        f'garf_compare_image_profile {p / "test043_projection_analog.mhd"} {filename1} -w 3 -s 75'
    )
    print(
        f'garf_compare_image_profile {p / "test043_projection_analog_high_stat.mhd"} {filename2} -w 3'
    )
    print(
        f'garf_compare_image_profile {p / "test043_projection_analog_high_stat.mhd"} {filename2} -w 3 -s 75'
    )

    utility.test_ok(is_ok)



from box import Box
from scipy.spatial.transform import Rotation
import logging
import numpy as np
import meshio

from ..base import GateObject, process_cls, create_gate_object_from_dict
from ..utility import g4_units
from ..exception import fatal
import opengate_core as g4
from ..decorators import requires_fatal

from .utility import ensure_is_g4_rotation, ensure_is_g4_translation, vec_np_as_g4


logger = logging.getLogger(__name__)


class SolidBase(GateObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.g4_solid = None

    def close(self):
        self.release_g4_references()
        super().close()

    def release_g4_references(self):
        self.g4_solid = None

    def __getstate__(self):
        return_dict = super().__getstate__()
        return_dict["g4_solid"] = None
        return return_dict

    @property
    def solid_info(self):
        """Computes the properties of the solid associated with this volume."""
        # Note: This method only works in derived classes which implement the build_solid method.
        solid = self.build_solid()
        if solid is None:
            fatal(
                f"Cannot compute solid info for this volume {self.name}. Unable to build the solid."
            )
        r = Box()
        r.cubic_volume = solid.GetCubicVolume()
        r.surface_area = solid.GetSurfaceArea()
        pMin = g4.G4ThreeVector()
        pMax = g4.G4ThreeVector()
        solid.BoundingLimits(pMin, pMax)
        r.bounding_limits = [pMin, pMax]
        return r

    @property
    def bounding_limits(self):
        """
        Return the min and max 3D points of the bounding box of the given volume
        """
        pMin, pMax = self.solid_info.bounding_limits
        return pMin, pMax

    @property
    def bounding_box_size(self):
        """
        Return the size of the bounding box of the given volume
        """
        pMin, pMax = self.bounding_limits
        return [pMax[0] - pMin[0], pMax[1] - pMin[1], pMax[2] - pMin[2]]

    # The construct_solid method is implemented here, but will only work with objects
    # of the derived classes which implement the build_solid method,
    # or which override the construct_solid method
    def construct_solid(self):
        """Attempts to build the solid according the build_solid() method either coming from a Solid mother class
        or implemented in a specific derived class.
        """
        # The solid can only be constructed once
        if self.g4_solid is None:
            self.g4_solid = self.build_solid()


class BooleanSolid(SolidBase):
    constructor_functions = {
        "intersect": g4.G4IntersectionSolid,
        "add": g4.G4UnionSolid,
        "subtract": g4.G4SubtractionSolid,
    }

    user_info_defaults = {
        "creator_volumes": (
            [None, None],
            {
                "doc": "A tuple of the two volumes which were combined by boolean operation to create this volume. "
                "This user info is set internally when applying a boolean operation "
                "and cannot be set by the user. ",
                "read_only": True,
            },
        ),
        "operation": ("none", {"doc": "FIXME"}),
        "rotation_boolean_operation": (
            Rotation.identity().as_matrix(),
            {"doc": "FIXME"},
        ),
        "translation_boolean_operation": ([0, 0, 0], {"doc": "FIXME"}),
    }

    def build_solid(self):
        """Overrides the method from the base class.
        It constructs the solid according to the logic of the G4 boolean volumes.
        """
        g4_rotation = ensure_is_g4_rotation(self.rotation_boolean_operation)
        g4_translation = ensure_is_g4_translation(self.translation_boolean_operation)

        # make sure creator volumes have their solids constructed
        for cv in self.creator_volumes:
            cv.construct_solid()

        return self.constructor_functions[self.operation](
            self.name,
            self.creator_volumes[0].g4_solid,
            self.creator_volumes[1].g4_solid,
            g4_rotation,
            g4_translation,
        )

    def from_dictionary(self, d):
        super().from_dictionary(d)
        try:
            creator_volumes = d["user_info"]["creator_volumes"]
        except KeyError:
            fatal(
                f"Error while populating object named {self.name}: "
                "The provided dictionary does not contain an entry 'creator_volumes'."
            )
        for i, cv in enumerate(creator_volumes):
            try:
                vol = self.volume_manager.volumes[cv["user_info"]["name"]]
            except KeyError:
                vol = create_gate_object_from_dict(cv)

            self.creator_volumes[i] = vol
            self.creator_volumes[i].from_dictionary(cv)


class BoxSolid(SolidBase):
    user_info_defaults = {
        "size": (
            [10 * g4_units.cm, 10 * g4_units.cm, 10 * g4_units.cm],
            {"doc": "3 component list of side lengths of the box."},
        )
    }

    def build_solid(self):
        return g4.G4Box(
            self.name, self.size[0] / 2.0, self.size[1] / 2.0, self.size[2] / 2.0
        )


class HexagonSolid(SolidBase):
    """
    This is the special case of a six-sided polyhedron.
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html

    """

    user_info_defaults = {
        "height": (
            5 * g4_units.cm,
            {"doc": "Height of the hexagon volume."},
        ),
        "radius": (
            0.15 * g4_units.cm,
            {"doc": "Radius from the center to corners."},
        ),
    }

    def build_solid(self):
        deg = g4_units.deg
        phi_start = 0 * deg
        phi_total = 360 * deg
        num_side = 6
        num_zplanes = 2
        zplane = [-self.height / 2, self.height / 2]
        radius_inner = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        radius_outer = [self.radius] * num_side

        return g4.G4Polyhedra(
            self.name,
            phi_start,
            phi_total,
            num_side,
            num_zplanes,
            zplane,
            radius_inner,
            radius_outer,
        )


class ConsSolid(SolidBase):
    """Cone section.
    http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html

    """

    user_info_defaults = {
        "rmin1": (
            5 * g4_units.mm,
            {"doc": "Inner radius at the lower end."},
        ),
        "rmin2": (
            20 * g4_units.mm,
            {"doc": "Inner radius at the upper end."},
        ),
        "rmax1": (
            10 * g4_units.mm,
            {"doc": "Outer radius at the lower end."},
        ),
        "rmax2": (
            25 * g4_units.mm,
            {"doc": "Outer radius at the upper end."},
        ),
        "dz": (40 * g4_units.mm, {"doc": "Half length in Z."}),
        "sphi": (
            0 * g4_units.deg,
            {"doc": "Starting angle of the segment in radians."},
        ),
        "dphi": (
            45 * g4_units.deg,
            {"doc": "The angle of the segment in radians."},
        ),
    }

    def build_solid(self):
        return g4.G4Cons(
            self.name,
            self.rmin1,
            self.rmax1,
            self.rmin2,
            self.rmax2,
            self.dz,
            self.sphi,
            self.dphi,
        )


class PolyhedraSolid(SolidBase):
    """
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html

    """

    user_info_defaults = {
        "phi_start": (
            0 * g4_units.deg,
            {"doc": "Initial Phi starting angle"},
        ),
        "phi_total": (
            360 * g4_units.deg,
            {"doc": "Total Phi angle"},
        ),
        "num_side": (6, {"doc": "Number of sides."}),
        "num_zplanes": (2, {"doc": "Number Z planes."}),
        "zplane": (
            [-2.5 * g4_units.cm, 2.5 * g4_units.cm],
            {
                "doc": "Position of Z planes. Should be a list with one position per plane."
            },
        ),
        "radius_inner": (
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            {
                "doc": "Tangent distance to inner surface. Should be a list with one distance per side."
            },
        ),
        "radius_outer": (
            [0.15 * g4_units.cm] * 6,
            {
                "doc": "Tangent distance to outer surface. Should be a list with one distance per side."
            },
        ),
    }

    def build_solid(self):
        return g4.G4Polyhedra(
            self.name,
            self.phi_start,
            self.phi_total,
            self.num_side,
            self.num_zplanes,
            self.zplane,
            self.radius_inner,
            self.radius_outer,
        )


class SphereSolid(SolidBase):
    user_info_defaults = {
        "rmin": (0, {"doc": "Inner radius (0 means solid sphere)."}),
        "rmax": (
            1 * g4_units.mm,
            {"doc": "Outer radius of the sphere."},
        ),
        "sphi": (0, {"doc": ""}),
        "dphi": (
            360 * g4_units.deg,
            {"doc": "Angular size of the sphere section around the rotation axis. "},
        ),
        "stheta": (0, {"doc": ""}),
        "dtheta": (180 * g4_units.deg, {"doc": ""}),
    }

    def build_solid(self):
        return g4.G4Sphere(
            self.name,
            self.rmin,
            self.rmax,
            self.sphi,
            self.dphi,
            self.stheta,
            self.dtheta,
        )


class EllipsoidSolid(SolidBase):
    user_info_defaults = {
        "xSemiAxis": (1 * g4_units.mm, {"doc": "Semiaxis in X"}),
        "ySemiAxis": (
            1 * g4_units.mm,
            {"doc": "Semiaxis in Y"},
        ),
        "zSemiAxis": (
            1 * g4_units.mm,
            {"doc": "Semiaxis in Z"},
        ),
        "zBottomCut": (
            0 * g4_units.mm,
            {"doc": "lower cut plane level, Z"},
        ),
        "zTopCut": (
            0 * g4_units.mm,
            {"doc": "upper cut plane level, Z"},
        ),
    }

    def build_solid(self):
        return g4.G4Ellipsoid(
            self.name,
            self.xSemiAxis,
            self.ySemiAxis,
            self.zSemiAxis,
            self.zBottomCut,
            self.zTopCut,
        )


class TrapSolid(SolidBase):
    """
    http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html

    """

    user_info_defaults = {
        "dx1": (
            30 * g4_units.mm,
            {"doc": "Half x length of the side at y=-pdy1 of the face at -pdz"},
        ),
        "dx2": (
            40 * g4_units.mm,
            {"doc": "Half x length of the side at y=+pdy1 of the face at -pdz"},
        ),
        "dy1": (40 * g4_units.mm, {"doc": "Half y length at -pdz"}),
        "dy2": (16 * g4_units.mm, {"doc": "Half y length at +pdz"}),
        "dx3": (
            10 * g4_units.mm,
            {"doc": "Half x length of the side at y=-pdy2 of the face at +pdz"},
        ),
        "dx4": (
            14 * g4_units.mm,
            {"doc": "Half x length of the side at y=+pdy2 of the face at +pdz"},
        ),
        "dz": (60 * g4_units.mm, {"doc": "Half z length"}),
        "theta": (
            20 * g4_units.deg,
            {
                "doc": "Polar angle of the line joining the centres of the faces at -/+pdz"
            },
        ),
        "phi": (
            5 * g4_units.deg,
            {
                "doc": "Azimuthal angle of the line joining the centre of the face at -pdz "
                "to the centre of the face at +pdz"
            },
        ),
        "alp1": (
            10 * g4_units.deg,
            {
                "doc": "Angle with respect to the y axis from the centre of the side (lower endcap)"
            },
        ),
        "alp2": (
            10 * g4_units.deg,
            {
                "doc": "Angle with respect to the y axis from the centre of the side (upper endcap)"
            },
        ),
    }

    def build_solid(self):
        return g4.G4Trap(
            self.name,
            self.dz,
            self.theta,
            self.phi,
            self.dy1,
            self.dx1,
            self.dx2,
            self.alp1,
            self.dy2,
            self.dx3,
            self.dx4,
            self.alp2,
        )


class TrdSolid(SolidBase):
    """
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html?highlight=g4trd

    dx1 Half-length along X at the surface positioned at -dz
    dx2 Half-length along X at the surface positioned at +dz
    dy1 Half-length along Y at the surface positioned at -dz
    dy2 Half-length along Y at the surface positioned at +dz
    zdz Half-length along Z axis

    """

    user_info_defaults = {
        "dx1": (
            30 * g4_units.mm,
            {"doc": "Half-length along X at the surface positioned at -dz"},
        ),
        "dx2": (
            10 * g4_units.mm,
            {"doc": "dx2 Half-length along X at the surface positioned at +dz"},
        ),
        "dy1": (
            40 * g4_units.mm,
            {"doc": "dy1 Half-length along Y at the surface positioned at -dz"},
        ),
        "dy2": (
            15 * g4_units.mm,
            {"doc": "dy2 Half-length along Y at the surface positioned at +dz"},
        ),
        "dz": (
            15 * g4_units.mm,
            {"doc": "Half-length along Z axis"},
        ),
    }

    def build_solid(self):
        return g4.G4Trd(self.name, self.dx1, self.dx2, self.dy1, self.dy2, self.dz)


class TubsSolid(SolidBase):
    """
    http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html

    """

    user_info_defaults = {
        "rmin": (30 * g4_units.mm, {"doc": "Inner radius"}),
        "rmax": (40 * g4_units.mm, {"doc": "Outer radius"}),
        "dz": (40 * g4_units.mm, {"doc": "Half length along Z"}),
        "sphi": (0 * g4_units.deg, {"doc": "Start angle phi"}),
        "dphi": (360 * g4_units.deg, {"doc": "Angle segment"}),
    }

    def build_solid(self):
        return g4.G4Tubs(self.name, self.rmin, self.rmax, self.dz, self.sphi, self.dphi)


class TesselatedSolid(SolidBase):
    """
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html?highlight=tesselated#tessellated-solids
    """

    file_name: str
    origin_at_cog: bool
    size_unit: float

    user_info_defaults = {
        "file_name": (
            "",
            {"doc": "Path and file name of the mesh file."},
        ),
        "origin_at_cog": (
            True,
            {"doc": "Translate the volume so its centre of gravity is at (0, 0, 0)."},
        ),
        "size_unit": (
            g4_units.mm,
            {"doc": "Unit to interpret data in the mesh file."},
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.g4_facets = None

    def close(self):
        # G4Facets are deleted by the destructor of G4TesselatedSolid,
        # so we need to release the reference to it on the python side.
        # self.g4_facets = None
        super().close()

    def build_solid(self):
        mm = g4_units.mm
        unit = self.size_unit

        mesh = meshio.read(self.file_name)
        points = mesh.points

        # translate the mesh to the center of gravity
        if self.origin_at_cog:
            centre = np.mean(points, axis=0)
            points = [point - centre for point in points]

        tessellated_solid = g4.G4TessellatedSolid(self.name)
        if "triangle" in mesh.cells_dict:
            for triangle in mesh.cells_dict["triangle"]:
                g4_facet = g4.G4TriangularFacet(
                    vec_np_as_g4(points[triangle[0]] * unit),
                    vec_np_as_g4(points[triangle[1]] * unit),
                    vec_np_as_g4(points[triangle[2]] * unit),
                    g4.G4FacetVertexType.ABSOLUTE,
                )
                tessellated_solid.AddFacet(g4_facet)

        if "quad" in mesh.cells_dict:
            for quad in mesh.cells_dict["quad"]:
                g4_facet = g4.G4QuadrangularFacet(
                    vec_np_as_g4(points[quad[0]] * unit),
                    vec_np_as_g4(points[quad[1]] * unit),
                    vec_np_as_g4(points[quad[2]] * unit),
                    vec_np_as_g4(points[quad[3]] * unit),
                    g4.G4FacetVertexType.ABSOLUTE,
                )
                tessellated_solid.AddFacet(g4_facet)

        # set the solid closed
        tessellated_solid.SetSolidClosed(True)
        logger.debug(
            f"Created tesselated volume '{self.name}'"
            f"with a volume of {tessellated_solid.GetCubicVolume() * unit**3 / mm**3} mm³"
        )

        return tessellated_solid


class ImageSolid(SolidBase):
    """Utility to handle the solids of an ImageVolume.
    It is not intended to be used stand-alone, but only as a base class of ImageVolume.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # These parameters are NOT user infos
        # They are derived from the ITK image in the ImageVolume class
        self.half_size_mm = None
        self.half_spacing = None

        self.g4_solid_x = None
        self.g4_solid_y = None
        self.g4_solid_z = None

        self.slice_xy = None
        self.slice_xz = None
        self.slice_yz = None

    def __getstate__(self):
        return_dict = super().__getstate__()
        return_dict["g4_solid_x"] = None
        return_dict["g4_solid_y"] = None
        return_dict["g4_solid_z"] = None
        return_dict["slice_xy"] = None
        return_dict["slice_xz"] = None
        return_dict["slice_yz"] = None
        return return_dict

    def close(self):
        self.release_g4_references()
        super().close()

    def release_g4_references(self):
        self.g4_solid_x = None
        self.g4_solid_y = None
        self.g4_solid_z = None
        self.slice_xy = None
        self.slice_xz = None
        self.slice_yz = None

    @requires_fatal("half_size_mm")
    @requires_fatal("half_spacing")
    def construct_solid(self):
        self.g4_solid_x = g4.G4Box(
            self.name + "_X",
            self.half_spacing[0],
            self.half_spacing[1],
            self.half_size_mm[2],
        )
        self.g4_solid_y = g4.G4Box(
            self.name + "_Y",
            self.half_size_mm[0],
            self.half_spacing[1],
            self.half_size_mm[2],
        )
        self.g4_solid_z = g4.G4Box(
            self.name + "_Z",
            self.half_spacing[0],
            self.half_spacing[1],
            self.half_spacing[2],
        )

        image_dict = {
            "name": self.name,
            "half_size_mm": self.half_size_mm,
            "size": (self.half_size_mm / self.half_spacing).astype(int),
        }
        self.g4_solid = g4.GateImageBox(image_dict)


process_cls(SolidBase)
process_cls(BooleanSolid)
process_cls(BoxSolid)
process_cls(HexagonSolid)
process_cls(ConsSolid)
process_cls(PolyhedraSolid)
process_cls(SphereSolid)
process_cls(EllipsoidSolid)
process_cls(TrapSolid)
process_cls(TrdSolid)
process_cls(TubsSolid)
process_cls(ImageSolid)
process_cls(TesselatedSolid)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test028_ge_nm670_spect_2_helpers as test028
from opengate.tests import utility
from pathlib import Path

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "", output_folder="test028_hits_mt"
    )

    # create the simulation
    sim = gate.Simulation()

    # main description
    test028.create_spect_simu(sim, paths, number_of_threads=3)

    # go
    sim.run()

    # check
    output_ref_folder = Path(str(paths.output_ref).replace("_mt", ""))
    output_ref_filename = "proj028_counts.mhd".replace("_mt", "")
    print(output_ref_folder)
    print(output_ref_filename)
    is_ok = test028.test_spect_root(sim, paths)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
from opengate.tests import utility
import matplotlib.pyplot as plt


def simulate(number_of_dynamic_parametrisations=0):
    paths = utility.get_default_test_paths(
        __file__, "gate_test029_volume_time_rotation", "test030"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.random_seed = 983456
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.second

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.translation = [1 * cm, 2 * cm, 3 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = "fake"
    waterbox.size = [20 * cm, 20 * cm, 20 * cm]
    waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
    waterbox.rotation = Rotation.from_euler("y", -20, degrees=True).as_matrix()
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.set_production_cut("world", "all", 700 * um)

    # default source for tests
    # the source is fixed at the center, only the volume will move
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 150 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 5 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 30 * Bq

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test030-edep.mhd"
    dose.attached_to = "waterbox"
    dose.size = [99, 99, 99]
    mm = gate.g4_units.mm
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose.translation = [2 * mm, 3 * mm, -2 * mm]
    dose.edep_uncertainty.active = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    # go
    sim.run(start_new_process=True)

    if number_of_dynamic_parametrisations > 1:
        # fake motion
        interval_length = 1 * sec / number_of_dynamic_parametrisations
        sim.run_timing_intervals = [
            (i * interval_length, (i + 1) * interval_length)
            for i in range(number_of_dynamic_parametrisations)
        ]
        # create a dynamic parametrisation which does actually not move anything
        # This is just to measure the run time overhead due to the call to BeginOfRunActionMasterThread()
        fake.add_dynamic_parametrisation(
            translation=number_of_dynamic_parametrisations * [fake.translation],
            rotation=number_of_dynamic_parametrisations * [fake.rotation],
        )

    # go
    sim.run(start_new_process=True)

    return stats.counts["duration"] / gate.g4_units["s"]


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test030")

    duration_static = simulate(0)
    n_dyn = [10, 100, 1000]
    durations = []
    for n in n_dyn:
        durations.append(simulate(n))

    print("********")
    print(f"The static simulation took {duration_static} sec")
    print(f"The dynamic simulations took: {duration_static} sec")
    for d, n in zip(durations, n_dyn):
        print(f"{d} sec for {n} dynamic parametrisations of translation and rotation")
    print("********")

    plt.figure()
    plt.scatter(n_dyn, durations, label="Dynamic simulation")
    plt.axhline(duration_static, label="Static simulation")
    plt.legend(loc="best")
    plt.xlabel("Number of dynamic parametrisations (translation, rotation)")
    plt.ylabel("Geant4 simulation run time in sec")
    plt.tight_layout()
    # plt.show()

    plt.savefig(paths.output / "run_time_dynamic_parametrisation.pdf")

    # FIXME: this is not really a test, always ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests.utility import *
from opengate.managers import Simulation
from opengate.actors.digitizers import *

if __name__ == "__main__":
    paths = get_default_test_paths(__file__, None, "test093")

    # units
    m = g4_units.m
    mm = g4_units.mm
    cm = g4_units.cm
    keV = g4_units.keV

    # simulation
    sim = Simulation()
    sim.progress_bar = True
    sim.visu = False
    sim.visu_type = "qt"
    sim.output_dir = paths.output
    sim.random_seed = 32145987

    # add a volume
    waterbox1 = sim.add_volume("Box", "Waterbox1")
    waterbox1.size = [20 * cm, 20 * cm, 10 * cm]
    waterbox1.translation = [0 * cm, 0 * cm, 0 * cm]
    waterbox1.material = "G4_WATER"

    # add a volume
    waterbox2 = sim.add_volume("Box", "Waterbox2")
    waterbox2.size = [20 * cm, 20 * cm, 10 * cm]
    waterbox2.translation = [0 * cm, 0 * cm, -10 * cm]
    waterbox2.material = "G4_WATER"

    # add a "detector"
    detector = sim.add_volume("Box", "detector")
    detector.size = [150 * cm, 150 * cm, 1 * cm]
    detector.translation = [0 * cm, 0 * cm, 20 * cm]
    detector.material = "G4_LEAD_OXIDE"

    # source 1
    source = sim.add_source("GenericSource", "low_energy")
    source.particle = "gamma"
    source.energy.mono = 50 * keV
    source.position.type = "sphere"
    source.position.radius = 20 * mm
    source.position.translation = [0 * cm, 0 * cm, -30 * cm]
    source.direction.type = "focused"
    source.direction.focus_point = [0 * cm, 0 * cm, -25 * cm]
    source.direction.momentum = [0, 0, -1]
    source.n = 50000

    # add a stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # needed for Rayl !
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"

    # phsp
    phsp = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phsp.attached_to = detector

    att1 = ProcessDefinedStepInVolumeAttribute(sim, "compt", waterbox1.name)
    att2 = ProcessDefinedStepInVolumeAttribute(sim, "compt", "world")
    att3 = ProcessDefinedStepInVolumeAttribute(sim, "Rayl", "world")
    att4 = ProcessDefinedStepInVolumeAttribute(sim, "Rayl", waterbox2.name)
    phsp.attributes = [
        "KineticEnergy",
        "PrePosition",
        "EventID",
        att1.name,
        att2.name,
        att3.name,
        att4.name,
    ]

    # not possible:
    is_ok = True
    try:
        att3.volume_name = "toto"
        att3.process_name = "toto"
        print(f"volume_name and process_name cannot be changed")
        is_ok = False
    except Exception as e:
        print("OK, cannot change the volume_name")

    phsp.output_filename = "phsp.root"
    phsp.steps_to_store = "all"

    # go
    sim.run()
    print(stats)

    # check
    ref_root = paths.output_ref / "phsp.root"
    test_root = paths.output / "phsp.root"
    print(phsp.attributes)
    keys = [att1.name, att2.name, att3.name, att4.name]
    n = len(keys)
    is_ok = (
        compare_root3(
            ref_root,
            test_root,
            "PhaseSpace",
            "PhaseSpace",
            keys1=keys,
            keys2=keys,
            tols=[0.07, 0.07, 0.05, 0.05],
            scalings1=[1] * n,
            scalings2=[1] * n,
            img=paths.output / "phsp.png",
        )
        and is_ok
    )

    test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
from test095_spect_helpers import *
import json


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, gate_folder="", output_folder="test095"
    )

    p = intevo.get_geometrical_parameters()
    collimators = p.collimators
    # collimators = ["lehr"]
    n = len(collimators)
    collimators.append(collimators[0])
    is_ok = True

    i = 0
    for colli_type in collimators:
        # rotate the gantry only for the last collimator
        if i == n:
            radius = 20 * gate.g4_units.cm
            rotation = 12
        else:
            radius = 0
            rotation = 0

        sim, fn = build_test_simu(paths, "intevo", colli_type, rotation, radius)

        # go
        sim.run(start_new_process=True)

        # compare the json files
        ref_json_file = paths.output_ref / fn
        print(ref_json_file)
        j_ref = json.load(open(ref_json_file))
        j_test = json.load(open(sim.user_hook_after_init_arg))
        added, removed, modified, same = utility.dict_compare(
            j_ref,
            j_test,
            tolerance=1e5,
            ignored_keys=[
                "copy_no",
            ],
        )
        b = len(added) == 0 and len(removed) == 0 and len(modified) == 0
        utility.print_test(b, f"Compare json volume info with reference")
        is_ok = is_ok and b
        i = i + 1

    utility.test_ok(is_ok)
